# coding: utf-8
"""Unit tests for reviewboard.reviews.views.CommentDiffFragmentsView."""

from __future__ import unicode_literals

import struct

from django.contrib.auth.models import User
from django.utils import six
from djblets.testing.decorators import add_fixtures

from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase


class CommentDiffFragmentsViewTests(TestCase):
    """Unit tests for reviewboard.reviews.views.CommentDiffFragmentsView."""

    fixtures = ['test_users', 'test_scmtools']

    def test_get_with_unpublished_review_request_not_owner(self):
        """Testing CommentDiffFragmentsView with unpublished review request and
        user is not the owner
        """
        user = User.objects.create_user(username='reviewer',
                                        password='reviewer',
                                        email='reviewer@example.com')

        review_request = self.create_review_request(create_repository=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request, user=user)
        comment1 = self.create_diff_comment(review, filediff)
        comment2 = self.create_diff_comment(review, filediff)
        review.publish()

        self.assertTrue(self.client.login(username='reviewer',
                                          password='reviewer'))

        self._get_fragments(review_request,
                            [comment1.pk, comment2.pk],
                            expected_status=403)

    def test_get_with_unpublished_review_request_owner(self):
        """Testing CommentDiffFragmentsView with unpublished review request and
        user is the owner
        """
        user = User.objects.create_user(username='test-user',
                                        password='test-user',
                                        email='user@example.com')

        review_request = self.create_review_request(create_repository=True,
                                                    submitter=user)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request, user=user)
        comment1 = self.create_diff_comment(review, filediff)
        comment2 = self.create_diff_comment(review, filediff)
        review.publish()

        self.assertTrue(self.client.login(username='test-user',
                                          password='test-user'))

        fragments = self._get_fragments(review_request,
                                        [comment1.pk, comment2.pk])
        self.assertEqual(len(fragments), 2)
        self.assertEqual(fragments[0][0], comment1.pk)
        self.assertEqual(fragments[1][0], comment2.pk)

    @add_fixtures(['test_site'])
    def test_get_with_published_review_request_local_site_access(self):
        """Testing CommentDiffFragmentsView with published review request on
        a Local Site the user has access to
        """
        user = User.objects.create_user(username='test-user',
                                        password='test-user',
                                        email='user@example.com')

        review_request = self.create_review_request(create_repository=True,
                                                    with_local_site=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request)
        comment1 = self.create_diff_comment(review, filediff)
        comment2 = self.create_diff_comment(review, filediff)
        review.publish()

        review_request.local_site.users.add(user)

        self.assertTrue(self.client.login(username='test-user',
                                          password='test-user'))

        fragments = self._get_fragments(review_request,
                                        [comment1.pk, comment2.pk],
                                        local_site_name='local-site-1')
        self.assertEqual(len(fragments), 2)
        self.assertEqual(fragments[0][0], comment1.pk)
        self.assertEqual(fragments[1][0], comment2.pk)

    @add_fixtures(['test_site'])
    def test_get_with_published_review_request_local_site_no_access(self):
        """Testing CommentDiffFragmentsView with published review request on
        a Local Site the user does not have access to
        """
        User.objects.create_user(username='test-user',
                                 password='test-user',
                                 email='user@example.com')

        review_request = self.create_review_request(create_repository=True,
                                                    with_local_site=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request)
        comment1 = self.create_diff_comment(review, filediff)
        comment2 = self.create_diff_comment(review, filediff)
        review.publish()

        self.assertTrue(self.client.login(username='test-user',
                                          password='test-user'))

        self._get_fragments(review_request,
                            [comment1.pk, comment2.pk],
                            local_site_name='local-site-1',
                            expected_status=403)

    def test_get_with_unicode(self):
        """Testing CommentDiffFragmentsView with Unicode content"""
        user = User.objects.create(username='reviewer')

        repository = self.create_repository(tool_name='Test')
        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(
            diffset,
            source_file='/data:áéíóú🔥',
            dest_file='/data:ÄËÏÖÜŸ',
            diff=(
                'diff --git a/data b/data\n'
                'index abcd123..abcd124 100644\n'
                '--- a/data\n'
                '+++ b/data\n'
                '@@ -1,1 +1,1 @@\n'
                '-áéíóú🔥\n'
                '+ÄËÏÖÜŸ\n'
            ).encode('utf-8'))

        review = self.create_review(review_request, user=user)
        comment1 = self.create_diff_comment(review, filediff)
        comment2 = self.create_diff_comment(review, filediff)
        review.publish()

        fragments = self._get_fragments(review_request,
                                        [comment1.pk, comment2.pk])
        self.assertEqual(len(fragments), 2)

        comment_id, html = fragments[0]
        self.assertEqual(comment_id, comment1.pk)
        self.assertTrue(html.startswith('<table class="sidebyside'))
        self.assertTrue(html.endswith('</table>'))
        self.assertIn('áéíóú🔥', html)

        comment_id, html = fragments[1]
        self.assertEqual(comment_id, comment2.pk)
        self.assertTrue(html.startswith('<table class="sidebyside'))
        self.assertTrue(html.endswith('</table>'))
        self.assertIn('ÄËÏÖÜŸ', html)

    def test_get_with_valid_comment_ids(self):
        """Testing CommentDiffFragmentsView with valid comment ID"""
        user = User.objects.create_user(username='reviewer',
                                        email='reviewer@example.com')

        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request, user=user)
        comment1 = self.create_diff_comment(review, filediff)
        comment2 = self.create_diff_comment(review, filediff)
        review.publish()

        fragments = self._get_fragments(review_request,
                                        [comment1.pk, comment2.pk])
        self.assertEqual(len(fragments), 2)
        self.assertEqual(fragments[0][0], comment1.pk)
        self.assertEqual(fragments[1][0], comment2.pk)

    def test_get_with_valid_and_invalid_comment_ids(self):
        """Testing CommentDiffFragmentsView with mix of valid comment IDs and
        comment IDs not found in database
        """
        user = User.objects.create_user(username='reviewer',
                                        email='reviewer@example.com')

        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request, user=user)
        comment = self.create_diff_comment(review, filediff)
        review.publish()

        fragments = self._get_fragments(review_request, [999, comment.pk])
        self.assertEqual(len(fragments), 1)
        self.assertEqual(fragments[0][0], comment.pk)

    def test_get_with_no_valid_comment_ids(self):
        """Testing CommentDiffFragmentsView with no valid comment IDs"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)

        self._get_fragments(review_request,
                            [100, 200, 300],
                            expected_status=404)

    def test_get_with_comment_ids_from_other_review_request(self):
        """Testing CommentDiffFragmentsView with comment ID from another review
        request
        """
        user = User.objects.create_user(username='reviewer',
                                        email='reviewer@example.com')

        # Create the first review request and review.
        review_request1 = self.create_review_request(create_repository=True,
                                                     publish=True)
        diffset = self.create_diffset(review_request1)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request1, user=user)
        comment1 = self.create_diff_comment(review, filediff)
        review.publish()

        # Create the second review request and review.
        review_request2 = self.create_review_request(create_repository=True,
                                                     publish=True)
        diffset = self.create_diffset(review_request2)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request2, user=user)
        comment2 = self.create_diff_comment(review, filediff)
        review.publish()

        fragments = self._get_fragments(review_request1,
                                        [comment1.pk, comment2.pk])
        self.assertEqual(len(fragments), 1)
        self.assertEqual(fragments[0][0], comment1.pk)

    def test_get_with_comment_ids_from_draft_review_owner(self):
        """Testing CommentDiffFragmentsView with comment ID from draft review,
        accessed by the review's owner
        """
        user = User.objects.create_user(username='reviewer',
                                        password='reviewer',
                                        email='reviewer@example.com')

        review_request1 = self.create_review_request(create_repository=True,
                                                     publish=True)
        diffset = self.create_diffset(review_request1)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request1, user=user)
        comment = self.create_diff_comment(review, filediff)

        self.assertTrue(self.client.login(username='reviewer',
                                          password='reviewer'))

        fragments = self._get_fragments(review_request1, [comment.pk])
        self.assertEqual(len(fragments), 1)
        self.assertEqual(fragments[0][0], comment.pk)

    def test_get_with_comment_ids_from_draft_review_not_owner(self):
        """Testing CommentDiffFragmentsView with comment ID from draft review,
        accessed by someone other than the review's owner
        """
        user = User.objects.create_user(username='reviewer',
                                        email='reviewer@example.com')

        review_request1 = self.create_review_request(create_repository=True,
                                                     publish=True)
        diffset = self.create_diffset(review_request1)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request1, user=user)
        comment = self.create_diff_comment(review, filediff)

        self._get_fragments(review_request1,
                            [comment.pk],
                            expected_status=404)

    def _get_fragments(self, review_request, comment_ids,
                       local_site_name=None, expected_status=200):
        """Load and return fragments from the server.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request the comments were made on.

            comment_ids (list of int):
                The list of comment IDs to load.

            local_site_name (unicode, optional):
                The name of the Local Site for the URL.

            expected_status (int, optional):
                The expected HTTP status code. By default, this is a
                successful 200.

        Returns:
            list of tuple:
            A list of ``(comment_id, html)`` from the parsed payload, if
            the status code was 200.
        """
        response = self.client.get(
            local_site_reverse(
                'diff-comment-fragments',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'comment_ids': ','.join(
                        six.text_type(comment_id)
                        for comment_id in comment_ids
                    ),
                },
                local_site_name=local_site_name))
        self.assertEqual(response.status_code, expected_status)

        if expected_status != 200:
            return None

        content = response.content
        self.assertIs(type(content), bytes)

        i = 0
        results = []

        while i < len(content):
            # Read the comment ID.
            comment_id = struct.unpack_from('<L', content, i)[0]
            i += 4

            # Read the length of the HTML.
            html_len = struct.unpack_from('<L', content, i)[0]
            i += 4

            # Read the HTML.
            html = content[i:i + html_len].decode('utf-8')
            i += html_len

            results.append((comment_id, html))

        return results



from __future__ import unicode_literals

from django.contrib.auth.models import User
from djblets.mail.utils import build_email_address_for_user
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.models import Profile
from reviewboard.notifications.email.utils import (
    build_recipients,
    get_email_addresses_for_group,
    recipients_to_addresses)
from reviewboard.reviews.models import Group
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class EmailUtilsTests(TestCase):
    """Testing e-mail utilities that do not send e-mails."""

    def test_recipients_to_addresses_with_string_address(self):
        """Testing generating addresses from recipients with string recipients
        """
        with self.assertRaises(AssertionError):
            recipients_to_addresses(['foo@example.com'])

    @add_fixtures(['test_users'])
    def test_recipients_to_addresses_with_users(self):
        """Testing generating addresses from recipients with user recipients
        """
        users = list(User.objects.filter(username__in=['doc', 'grumpy']))

        addresses = recipients_to_addresses(users)
        self.assertEqual(len(addresses), 2)

        expected_addresses = set(
            build_email_address_for_user(u)
            for u in users
        )

        self.assertEqual(addresses, expected_addresses)

    def test_recipients_to_addresses_with_groups_single_mailinglist(self):
        """Testing generating addresses from recipients that are groups with a
        single mailing list address
        """
        groups = [
            Group(name='group1', display_name='Group One',
                  mailing_list='group1@example.com'),
            Group(name='group2', display_name='Group Two',
                  mailing_list='group2@example.com'),
        ]

        addresses = recipients_to_addresses(groups)
        self.assertEqual(len(addresses), 2)

        expected_addresses = set(sum(
            (
                get_email_addresses_for_group(group)
                for group in groups
            ),
            []))

        self.assertEqual(addresses, expected_addresses)

    def test_recipients_to_addresses_with_groups_many_mailinglist(self):
        """Testing generating addresses from recipients that are groups with
        multiple mailing list addresses
        """
        groups = [
            Group(name='group1', display_name='Group One',
                  mailing_list='group1a@example.com,group1b@example.com'),
            Group(name='group2', display_name='Group Two',
                  mailing_list='group2a@example.com,group2b@example.com'),
        ]

        addresses = recipients_to_addresses(groups)
        self.assertEqual(len(addresses), 4)

        expected_addresses = set(sum(
            (
                get_email_addresses_for_group(group)
                for group in groups
            ),
            []))

        self.assertEqual(addresses, expected_addresses)

    @add_fixtures(['test_users'])
    def test_recipients_to_addresses_with_groups_and_users(self):
        """Testing generating addresses from recipients that are users and
        groups with mailing list addresses
        """
        groups = [
            Group(name='group1', display_name='Group One',
                  mailing_list='group1@example.com'),
            Group(name='group2', display_name='Group Two',
                  mailing_list='group2@example.com'),
        ]

        users = list(User.objects.filter(username__in=['doc', 'grumpy']).all())

        addresses = recipients_to_addresses(groups + users)
        self.assertEqual(len(addresses), 4)

        user_addresses = [
            build_email_address_for_user(u)
            for u in users
        ]

        group_addresses = sum(
            (
                get_email_addresses_for_group(group)
                for group in groups
            ),
            [])

        self.assertEqual(addresses,
                         set(user_addresses + group_addresses))

    def test_recipients_to_addresses_with_groups_with_members(self):
        """Testing generating addresses from recipients that are groups with
        no mailing list addresses
        """
        group1 = Group.objects.create(name='group1')
        group2 = Group.objects.create(name='group2')

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='user2', first_name='User',
                                         last_name='Two',
                                         email='user2@example.com')

        group1.users = [user1]
        group2.users = [user2]

        addresses = recipients_to_addresses([group1, group2])

        expected_addresses = set([
            build_email_address_for_user(user1),
            build_email_address_for_user(user2),
        ])

        self.assertEqual(addresses, expected_addresses)

    def test_recipients_to_addresses_with_groups_local_site(self):
        """Testing generating addresses from recipients that are groups in
        local sites
        """
        local_site1 = LocalSite.objects.create(name='local-site1')
        local_site2 = LocalSite.objects.create(name='local-site2')

        group1 = Group.objects.create(name='group1', local_site=local_site1)
        group2 = Group.objects.create(name='group2', local_site=local_site2)

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='user2', first_name='User',
                                         last_name='Two',
                                         email='user2@example.com')

        local_site1.users = [user1]

        group1.users = [user1]
        group2.users = [user2]

        addresses = recipients_to_addresses([group1, group2])
        self.assertEqual(len(addresses), 1)
        self.assertEqual(addresses, set([build_email_address_for_user(user1)]))

    def test_recipients_to_addresses_with_groups_inactive_members(self):
        """Testing generating addresses form recipients that are groups with
        inactive members
        """
        group1 = self.create_review_group('group1')
        group2 = self.create_review_group('group2')

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create(username='user2', first_name='User',
                                    last_name='Two', is_active=False,
                                    email='user1@example.com')

        group1.users = [user1]
        group2.users = [user2]

        addresses = recipients_to_addresses([group1, group2])
        self.assertEqual(len(addresses), 1)
        self.assertEqual(addresses, set([build_email_address_for_user(user1)]))

    def test_recipients_to_addresses_groups_local_site_inactive_members(self):
        """Testing generating addresses from recipients that are groups in
        local sites that have inactive members
        """
        local_site1 = LocalSite.objects.create(name='local-site1')
        local_site2 = LocalSite.objects.create(name='local-site2')

        group1 = self.create_review_group('group1', local_site=local_site1)
        group2 = self.create_review_group('group2', local_site=local_site2)

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create(username='user2', first_name='User',
                                    last_name='Two', is_active=False,
                                    email='user2@example.com')

        local_site1.users = [user1]
        local_site2.users = [user2]

        group1.users = [user1]
        group2.users = [user2]

        addresses = recipients_to_addresses([group1, group2])
        self.assertEqual(len(addresses), 1)
        self.assertEqual(addresses, set([build_email_address_for_user(user1)]))

    @add_fixtures(['test_users'])
    def test_build_recipients_user_receive_email(self):
        """Testing building recipients for a review request where the user
        wants to receive e-mail
        """
        review_request = self.create_review_request()
        submitter = review_request.submitter

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([submitter]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_user_not_receive_email(self):
        """Testing building recipients for a review request where the user
        does not want to receive e-mail
        """
        review_request = self.create_review_request()
        submitter = review_request.submitter

        profile = submitter.get_profile()
        profile.should_send_email = False
        profile.save()

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(len(to), 0)
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_user_not_receive_own_email(self):
        """Testing building recipients for a review request where the user
        does not want to receive e-mail about their updates
        """
        review_request = self.create_review_request()
        submitter = review_request.submitter

        profile = submitter.get_profile()
        profile.should_send_own_updates = False
        profile.save()

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(len(to), 0)
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_target_people_not_receive_own_email(self):
        """Testing building recipieints for a review request where the
        submitter is a reviewer and doesn't want to receive e-mail about their
        updates
        """
        review_request = self.create_review_request()
        submitter = review_request.submitter

        review_request.target_people = [submitter]

        profile = submitter.get_profile()
        profile.should_send_own_updates = False
        profile.save()

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(len(to), 0)
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_extra_recipient_user_not_receive_own_email(self):
        """Testing building recipients for a review request where the
        submitter is a reviewer and doesn't want to receive e-mail about their
        updates
        """
        review_request = self.create_review_request()
        submitter = review_request.submitter

        profile = submitter.get_profile()
        profile.should_send_own_updates = False
        profile.save()

        to, cc = build_recipients(submitter, review_request, [submitter])

        self.assertEqual(len(to), 0)
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_target_people_and_groups(self):
        """Testing building recipients for a review request where there are
        target users and groups
        """
        group = self.create_review_group()
        user = User.objects.get(username='grumpy')

        review_request = self.create_review_request()
        review_request.target_people = [user]
        review_request.target_groups = [group]

        submitter = review_request.submitter

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([user]))
        self.assertEqual(cc, set([submitter, group]))

    @add_fixtures(['test_users'])
    def test_build_recipients_target_people_inactive_and_groups(self):
        """Testing building recipients for a review request where there are
        target groups and inactive target users
        """
        group = self.create_review_group()
        user = User.objects.create(username='user', first_name='User',
                                   last_name='Foo', is_active=False,
                                   email='user@example.com')

        review_request = self.create_review_request()
        review_request.target_people = [user]
        review_request.target_groups = [group]

        submitter = review_request.submitter

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([submitter, group]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_target_groups(self):
        """Testing build recipients for a review request where there are target
        groups
        """
        group1 = self.create_review_group('group1')
        group2 = self.create_review_group('group2')

        review_request = self.create_review_request()
        review_request.target_groups = [group1, group2]
        submitter = review_request.submitter

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(len(to), 3)
        self.assertEqual(to, set([submitter, group1, group2]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_target_people(self):
        """Testing building recipients for a review request with target people
        """
        review_request = self.create_review_request()
        submitter = review_request.submitter

        grumpy = User.objects.get(username='grumpy')
        review_request.target_people = [grumpy]

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([grumpy]))
        self.assertEqual(cc, set([submitter]))

    @add_fixtures(['test_users'])
    def test_build_recipients_target_people_inactive(self):
        """Testing building recipients for a review request with target people
        who are inactive
        """
        review_request = self.create_review_request()
        submitter = review_request.submitter

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create(username='user2', first_name='User',
                                    last_name='Two', email='user2@example.com',
                                    is_active=False)

        review_request.target_people = [user1, user2]

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([user1]))
        self.assertEqual(cc, set([submitter]))

    @add_fixtures(['test_users'])
    def test_build_recipients_target_people_no_email(self):
        """Testing building recipients for a review request with target people
        who don't receive e-mail
        """
        review_request = self.create_review_request()
        submitter = review_request.submitter

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='user2', first_name='User',
                                         last_name='Two',
                                         email='user2@example.com')

        Profile.objects.create(user=user2, should_send_email=False)

        review_request.target_people = [user1, user2]

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([user1]))
        self.assertEqual(cc, set([submitter]))

    @add_fixtures(['test_users'])
    def test_build_recipients_target_people_local_site(self):
        """Testing building recipients for a review request where the target
        people are in local sites
        """
        local_site = LocalSite.objects.create(name=self.local_site_name)

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='user2', first_name='User',
                                         last_name='Two',
                                         email='user2@example.com')

        local_site.users = [user1]

        review_request = self.create_review_request(with_local_site=True)
        review_request.target_people = [user1, user2]

        submitter = review_request.submitter

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([user1]))
        self.assertEqual(cc, set([submitter]))

    @add_fixtures(['test_users'])
    def test_build_recipients_target_people_local_site_inactive(self):
        """Testing building recipients for a review request where the target
        people are in local sites and are inactive
        """
        local_site = LocalSite.objects.create(name=self.local_site_name)

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create(username='user2', first_name='User',
                                    last_name='Two', is_active=False,
                                    email='user2@example.com')

        local_site.users = [user1, user2]

        review_request = self.create_review_request(with_local_site=True)
        review_request.target_people = [user1, user2]

        submitter = review_request.submitter

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([user1]))
        self.assertEqual(cc, set([submitter]))

    @add_fixtures(['test_users'])
    def test_build_recipients_target_people_local_site_no_email(self):
        """Testing building recipients for a review request where the target
        people are in local sites don't receieve e-mail
        """
        local_site = LocalSite.objects.create(name=self.local_site_name)

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='user2', first_name='User',
                                         last_name='Two',
                                         email='user2@exmaple.com')

        Profile.objects.create(user=user2,
                               should_send_email=False)

        local_site.users = [user1, user2]

        review_request = self.create_review_request(with_local_site=True)
        review_request.target_people = [user1, user2]

        submitter = review_request.submitter

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([user1]))
        self.assertEqual(cc, set([submitter]))

    @add_fixtures(['test_users'])
    def test_build_recipients_limit_to(self):
        """Testing building recipients with a limited recipients list"""
        dopey = User.objects.get(username='dopey')
        grumpy = User.objects.get(username='grumpy')
        group = self.create_review_group()

        review_request = self.create_review_request()
        submitter = review_request.submitter

        review_request.target_people = [dopey]
        review_request.target_groups = [group]

        to, cc = build_recipients(submitter, review_request,
                                  limit_recipients_to=[grumpy])

        self.assertEqual(to, set([submitter, grumpy]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_limit_to_inactive(self):
        """Testing building recipients with a limited recipients list that
        contains inactive users
        """
        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create(username='user2', first_name='User',
                                    last_name='Two', email='user2@exmaple.com',
                                    is_active=False)

        review_request = self.create_review_request()
        submitter = review_request.submitter

        to, cc = build_recipients(submitter, review_request,
                                  limit_recipients_to=[user1, user2])

        self.assertEqual(to, set([submitter, user1]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_limit_to_local_site(self):
        """Testing building recipients with a limited recipients list that
        contains users in local sites
        """
        local_site1 = LocalSite.objects.create(name='local-site1')
        local_site2 = LocalSite.objects.create(name='local-site2')

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='user2', first_name='User',
                                         last_name='Two',
                                         email='user2@exmaple.com')

        local_site1.users = [user1]
        local_site2.users = [user2]

        review_request = self.create_review_request(local_site=local_site1)
        submitter = review_request.submitter

        to, cc = build_recipients(submitter, review_request,
                                  limit_recipients_to=[user1, user2])

        self.assertEqual(to, set([submitter, user1]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_extra_recipients(self):
        """Testing building recipients with an extra recipients list"""
        review_request = self.create_review_request()
        submitter = review_request.submitter

        grumpy = User.objects.get(username='grumpy')

        to, cc = build_recipients(submitter, review_request,
                                  extra_recipients=[grumpy])

        self.assertEqual(to, set([submitter, grumpy]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_extra_recipients_inactive(self):
        """Testing building recipients with an extra recipients list that
        contains inactive users
        """
        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create(username='user2', first_name='User',
                                    last_name='Two', email='user2@exmaple.com',
                                    is_active=False)

        review_request = self.create_review_request()
        submitter = review_request.submitter

        to, cc = build_recipients(submitter, review_request,
                                  extra_recipients=[user1, user2])

        self.assertEqual(to, set([submitter, user1]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_extra_recipients_local_site(self):
        """Testing building recipients with an extra recipients list that
        contains users in local sites
        """
        local_site1 = LocalSite.objects.create(name='local-site1')
        local_site2 = LocalSite.objects.create(name='local-site2')

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='user2', first_name='User',
                                         last_name='Two',
                                         email='user2@exmaple.com')

        local_site1.users = [user1]
        local_site2.users = [user2]

        review_request = self.create_review_request(local_site=local_site1)
        submitter = review_request.submitter

        to, cc = build_recipients(submitter, review_request,
                                  extra_recipients=[user1, user2])

        self.assertEqual(to, set([submitter, user1]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_extra_recipients_and_limit_to(self):
        """Testing building recipients with an extra recipients list and
        a limited recipients list
        """
        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='user2', first_name='User',
                                         last_name='Two',
                                         email='user2@exmaple.com')
        user3 = User.objects.create_user(username='user3', first_name='User',
                                         last_name='Three',
                                         email='user3@exmaple.com')

        group = self.create_review_group()

        review_request = self.create_review_request()
        submitter = review_request.submitter
        review_request.target_people = [user3]
        review_request.target_groups = [group]

        to, cc = build_recipients(submitter, review_request,
                                  extra_recipients=[user1],
                                  limit_recipients_to=[user2])

        self.assertEqual(to, set([submitter, user2]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_extra_recipients_and_limit_to_inactive(self):
        """Testing building recipients with an extra recipients list and a
        limited recipients list that contains inactive users
        """
        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create(username='user2', first_name='User',
                                    last_name='Two', email='user2@exmaple.com',
                                    is_active=False)
        user3 = User.objects.create_user(username='user3', first_name='User',
                                         last_name='Three',
                                         email='user3@exmaple.com')

        group = self.create_review_group()

        review_request = self.create_review_request()
        submitter = review_request.submitter
        review_request.target_people = [user3]
        review_request.target_groups = [group]

        to, cc = build_recipients(submitter, review_request,
                                  extra_recipients=[user1],
                                  limit_recipients_to=[user2])

        self.assertEqual(to, set([submitter]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_extra_recipients_and_limit_to_local_site(self):
        """Testing building recipients with an extra recipients list and a
        limited recipients list that contains users in local sites
        """
        local_site1 = LocalSite.objects.create(name='local-site1')
        local_site2 = LocalSite.objects.create(name='local-site2')

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='user2', first_name='User',
                                         last_name='Two',
                                         email='user2@exmaple.com')
        user3 = User.objects.create_user(username='user3', first_name='User',
                                         last_name='Three',
                                         email='user3@exmaple.com')

        local_site1.users = [user1, user3]
        local_site2.users = [user2]

        group = self.create_review_group()

        review_request = self.create_review_request(local_site=local_site1)
        submitter = review_request.submitter
        review_request.target_people = [user3]
        review_request.target_groups = [group]

        to, cc = build_recipients(submitter, review_request,
                                  extra_recipients=[user1],
                                  limit_recipients_to=[user2])

        self.assertEqual(to, set([submitter]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_starred(self):
        """Testing building recipients where the review request has been
        starred by a user
        """
        review_request = self.create_review_request()
        submitter = review_request.submitter

        grumpy = User.objects.get(username='grumpy')
        profile = grumpy.get_profile()
        profile.starred_review_requests = [review_request]
        profile.save()

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([submitter, grumpy]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_starred_inactive(self):
        """Testing building recipients where the review request has been
        starred by users that may be inactive
        """
        review_request = self.create_review_request()
        submitter = review_request.submitter

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create(username='user2', first_name='User',
                                    last_name='Two', email='user@exmaple.com',
                                    is_active=False)
        profile1 = Profile.objects.create(user=user1)
        profile1.starred_review_requests = [review_request]

        profile2 = Profile.objects.create(user=user2)
        profile2.starred_review_requests = [review_request]

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([submitter, user1]))
        self.assertEqual(len(cc), 0)

    @add_fixtures(['test_users'])
    def test_build_recipients_starred_local_site(self):
        """Testing building recipients where the review request has been
        starred by users that are in local sites
        """
        local_site1 = LocalSite.objects.create(name='local-site1')
        local_site2 = LocalSite.objects.create(name='local-site2')

        review_request = self.create_review_request(local_site=local_site1)
        submitter = review_request.submitter

        user1 = User.objects.create_user(username='user1', first_name='User',
                                         last_name='One',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='user2', first_name='User',
                                         last_name='Two',
                                         email='user@exmaple.com')

        local_site1.users = [user1]
        local_site2.users = [user2]

        profile1 = Profile.objects.create(user=user1)
        profile1.starred_review_requests = [review_request]

        profile2 = Profile.objects.create(user=user2)
        profile2.starred_review_requests = [review_request]

        to, cc = build_recipients(submitter, review_request)

        self.assertEqual(to, set([submitter, user1]))
        self.assertEqual(len(cc), 0)



"""Command line tool for helping develop extensions.

This tool, :command:`rbext`, currently provides the ability to easily run
extension-provided unit tests. In the future, it'll also help with other
development tasks, and with updating the Package Store with the latest versions
of an extension.
"""

from __future__ import print_function, unicode_literals

import argparse
import logging
import os
import re
import sys
from textwrap import dedent

os.environ.setdefault(b'DJANGO_SETTINGS_MODULE', b'reviewboard.settings')

rbext_dir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, os.path.join(rbext_dir, 'conf', 'rbext'))

import pkg_resources
from django.utils.translation import ugettext_lazy as _, ugettext

from reviewboard import get_manual_url


# NOTE: We want to include Django-based modules as late as possible, in order
#       to allow extension-provided settings to apply.


class BaseCommand(object):
    """Base class for a command."""

    #: The name of a command.
    #:
    #: This is what the user will type on the command line after ``rbext``.
    name = None

    #: The summary of the command.
    #:
    #: This will be shown in the help output.
    help_summary = 'Run unit tests for an extension.'

    def add_options(self, parser):
        """Add any command-specific options to the parser.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser.
        """
        pass

    def run(self, options):
        """Run the command.

        This will perform common setup work and then hand things off to
        :py:meth:`main`.

        Args:
            options (argparse.Namespace):
                Options set from the arguments.
        """
        if options.settings_file:
            sys.path.insert(
                0,
                os.path.abspath(os.path.dirname(options.settings_file)))

        return self.main(options)

    def main(self, options):
        """Perform the main operations for the command.

        Args:
            options (argparse.Namespace):
                Options set from the arguments.

        Returns:
            int:
            The command's exit code.
        """
        raise NotImplementedError

    def error(self, msg):
        """Display a fatal error to the user and exit.

        Args:
            msg (unicode):
                The message to display.

        Raises:
            django.core.management.CommandError:
            The resulting error.
        """
        from django.core.management import CommandError

        raise CommandError(msg)


class TestCommand(BaseCommand):
    """A command that runs an extension's test suite."""

    name = 'test'
    help_summary = 'Run unit tests for an extension.'

    def add_options(self, parser):
        """Add command line arguments for running tests.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser.
        """
        parser.add_argument(
            '--tree-root',
            metavar='PATH',
            default=os.getcwd(),
            help='The path to the root of the source tree.')
        parser.add_argument(
            '-m',
            '--module',
            action='append',
            metavar='MODULE_NAME',
            dest='module_names',
            required=True,
            help='The name(s) of the extension module(s) to test. For '
                 'example, if your tests are in "myextension.tests", you '
                 'might want to use "myextension".')

        # Note that we never actually handle this argument anywhere here.
        # This is really just to satisfy the parser. The test runner itself
        # handles it directly.
        parser.add_argument(
            '--with-coverage',
            action='store_true',
            default=False,
            help='Generate a code coverage report for the tests.')

        parser.add_argument(
            'tests',
            metavar='TEST',
            nargs='*',
            help='Specific tests to run. This can be in the form of '
                 'mypackage.mymodule, mypackage.mymodule:TestsClass, or '
                 'mypackage.mymodule:TestsClass.test_method.')

    def main(self, options):
        """Main function for running unit tests for the extension.

        Args:
            options (argparse.Namespace):
                Options set from the arguments.

        Returns:
            int:
            The command's exit code.
        """
        os.environ[b'RB_TEST_MODULES'] = b','.join(
            module_name.encode('utf-8')
            for module_name in options.module_names
        )

        os.chdir(options.tree_root)
        os.environ[b'RB_RUNNING_TESTS'] = b'1'

        from reviewboard.test import RBTestRunner

        test_runner = RBTestRunner(test_packages=options.module_names,
                                   cover_packages=options.module_names,
                                   verbosity=1)
        failures = test_runner.run_tests(options.tests)

        if failures:
            return 1
        else:
            return 0


class CreateCommand(BaseCommand):
    """A command for creating a new extension package."""

    name = 'create'
    help_summary = _('Create a new extension source tree.')

    def add_options(self, parser):
        """Add command line arguments for creating an extension.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser.
        """
        parser.add_argument(
            '--name',
            required=True,
            help=_('The human-readable name for the extension. This is '
                   'required.'))
        parser.add_argument(
            '--class-name',
            default=None,
            help=_('The class name for the extension (generally in CamelCase '
                   'form, without spaces). If not provided, this will be '
                   'based on the extension name.'))
        parser.add_argument(
            '--package-name',
            default=None,
            help=_('The name of the package (using alphanumeric  ). '
                   'If not provided, this will be based on the exension '
                   'name.'))
        parser.add_argument(
            '--package-version',
            default='1.0',
            help=_('The version for your extension and package.'))
        parser.add_argument(
            '--summary',
            default=None,
            help=_('A one-line summary of the extension.'))
        parser.add_argument(
            '--description',
            default=None,
            help=_('A short description of the extension.'))
        parser.add_argument(
            '--author-name',
            default=None,
            help=_('The name of the author for the package and extension '
                   'metadata. This can be a company name.'))
        parser.add_argument(
            '--author-email',
            default=None,
            help=_('The e-mail address of the author for the package and '
                   'extension metadata.'))
        parser.add_argument(
            '--enable-configuration',
            action='store_true',
            default=False,
            help=_('Whether to enable a Configure button and view for the '
                   'extension.'))
        parser.add_argument(
            '--enable-static-media',
            action='store_true',
            default=False,
            help=_('Whether to enable static media files for the package.'))

    def main(self, options):
        """Main function for creating an extension.

        Args:
            options (argparse.Namesapce):
                Options set from the arguments.

        Returns:
            int:
            The comamnd's exit code.
        """
        self._process_options(options)

        name = options.name
        package_name = options.package_name
        summary = options.summary
        description = options.description
        class_name = options.class_name
        configurable = options.enable_configuration
        enable_static_media = options.enable_static_media

        # Create the directory hierarchy.
        root_dir = package_name

        if os.path.exists(root_dir):
            self.error(
                ugettext('There\'s already a directory named "%s". You must '
                         'remove it before you can create a new extension '
                         'there.')
                % root_dir)

        ext_dir = os.path.join(root_dir, package_name)
        static_dir = os.path.join(ext_dir, 'static')
        templates_dir = os.path.join(ext_dir, 'templates')

        for path in (root_dir, ext_dir):
            os.mkdir(path, 0o755)

        if enable_static_media:
            os.mkdir(static_dir, 0o755)

            for path in ('css', 'js', 'images'):
                os.mkdir(os.path.join(static_dir, path))

        # Create the packaging files.
        self._write_file(
            os.path.join(root_dir, 'README.rst'),
            self._create_readme(name=name,
                                summary=summary,
                                description=description))

        self._write_file(
            os.path.join(root_dir, 'MANIFEST.in'),
            self._create_manifest(static_dir=static_dir,
                                  templates_dir=templates_dir))

        self._write_file(
            os.path.join(root_dir, 'setup.py'),
            self._create_setup_py(package_name=package_name,
                                  version=options.package_version,
                                  summary=summary,
                                  author=options.author_name,
                                  author_email=options.author_email,
                                  class_name=class_name),
            mode=0o755)

        # Create the extension source files.
        self._write_file(os.path.join(ext_dir, '__init__.py'), '')

        self._write_file(
            os.path.join(ext_dir, 'extension.py'),
            self._create_extension_py(
                name=name,
                package_name=package_name,
                class_name=class_name,
                summary=summary,
                configurable=configurable,
                has_static_media=enable_static_media))

        if configurable:
            form_class_name = '%sForm' % class_name

            self._write_file(
                os.path.join(ext_dir, 'admin_urls.py'),
                self._create_admin_urls_py(
                    package_name=package_name,
                    class_name=class_name,
                    form_class_name=form_class_name))

            self._write_file(
                os.path.join(ext_dir, 'forms.py'),
                self._create_forms_py(form_class_name=form_class_name))

        # We're done!
        print('Generated a new extension in %s' % os.path.abspath(root_dir))
        print()
        print('For information on writing your extension, see')
        print('%sextending/' % get_manual_url())

        return 0

    def _process_options(self, options):
        """Process and normalize any provided options.

        This will attempt to provide suitable defaults for missing parameters,
        adn to check that others are valid.

        Args:
            options (argparse.Namesapce):
                Options set from the arguments.
        """
        name = options.name.strip()
        package_name = options.package_name
        class_name = options.class_name

        if not package_name:
            package_name = self._normalize_package_name(name)
            print(ugettext('Using "%s" as the package name.') % package_name)
        else:
            package_name = package_name.strip()

            if not re.match(r'[A-Za-z][A-Za-z0-9._-]*', package_name):
                self.error(
                    ugettext('"%s" is not a valid package name. Try '
                             '--package-name="%s"')
                    % (package_name,
                       self._normalize_package_name(package_name)))

        if not class_name:
            class_name = self._normalize_class_name(name)
            print(ugettext('Using "%s" as the extension class name.')
                  % class_name)
        else:
            class_name = class_name.strip()

            if not re.match(r'[A-Za-z][A-Za-z0-9_]+Extension$', class_name):
                self.error(
                    ugettext('"%s" is not a valid class name. Try '
                             '--class-name="%s"')
                    % (package_name,
                       self._normalize_class_name(class_name)))

        options.name = name
        options.package_name = package_name
        options.class_name = class_name

    def _normalize_package_name(self, name):
        """Normalize a package name.

        This will ensure the package name is in a suitable format, replacing
        any invalid characters or dashes with ``_``, and converting it to
        lowercase.

        Args:
            name (unicode):
                The name of the package to normalize.

        Returns:
            unicode:
            The normalized name.
        """
        return pkg_resources.safe_name(name).replace('-', '_').lower()

    def _normalize_class_name(self, name):
        """Normalize a class name.

        This will ensure the class name is in a suitable format, converting it
        to CamelCase and adding an "Extension" at the end if needed.

        Args:
            name (unicode):
                The name of the class to normalize.

        Returns:
            unicode:
            The normalized class name.
        """
        class_name = ''.join(
            word.capitalize()
            for word in re.sub('[^A-Za-z0-9]+', ' ', name).split()
        )

        if not class_name.endswith('Extension'):
            class_name += 'Extension'

        return class_name

    def _write_file(self, filename, content, mode=None):
        """Write content to a file.

        This will create the file and write the provided content. The content
        will be stripped and dedented, with a trailing newline added, allowing
        the generation code to make use of multi-line strings.

        Args:
            path (unicode):
                The path of the file to write.

            content (unicode):
                The content to write.

            mode (int):
                The optional permissions mode to set for the file.
        """
        with open(filename, 'w') as fp:
            fp.write(dedent(content).strip())
            fp.write('\n')

        if mode is not None:
            os.chmod(filename, mode)

    def _create_readme(self, name, summary, description):
        """Create the content for a README.rst file.

        Args:
            name (unicode):
                The extension's name.

            summary (unicode):
                The extension's summary.

            description (unicode):
                A description of the extension.

        Returns:
            unicode:
            The resulting content for the file.
        """
        return """
            %(header_bar)s
            %(header)s
            %(header_bar)s

            %(content)s
        """ % {
            'header': name,
            'header_bar': '=' * len(name),
            'content': '\n\n'.join(
                content
                for content in (summary, description)
                if content
            ) or ugettext('Describe your extension.'),
        }

    def _create_manifest(self, templates_dir, static_dir):
        """Create the content for a MANIFEST.in file.

        Args:
            templates_dir (unicode):
                The relative path to the templates directory.

            static_dir (unicode):
                The relative path to the static media directory.

        Returns:
            unicode:
            The resulting content for the file.
        """
        return """
            graft %(templates_dir)s
            graft %(static_dir)s

            include COPYING
            include INSTALL
            include README.md
            include *-requirements.txt

            global-exclude .*.sw[op] *.py[co] __pycache__ .DS_Store .noseids
        """ % {
            'templates_dir': templates_dir,
            'static_dir': static_dir,
        }

    def _create_setup_py(self, package_name, version, summary, author,
                         author_email, class_name):
        """Create the content for a setup.py file.

        Args:
            package_name (unicode):
                The name of the package.

            version (unicode):
                The version of the package.

            summary (unicode):
                A summary of the package.

            author (unicode):
                The name of the author of the extension.

            author_email (unicode):
                The e-mail address of the author of the extension.

            class_name (unicode):
                The name of the extension class.

        Returns:
            unicode:
            The resulting content for the file.
        """
        return """
            #!/usr/bin/env python

            from __future__ import unicode_literals

            from reviewboard.extensions.packaging import setup
            from setuptools import find_packages


            setup(
                name='%(package_name)s',
                version='%(version)s',
                description=%(description)s,
                author=%(author)s,
                author_email=%(author_email)s,
                packages=find_packages(),
                install_requires=[
                    # Your package dependencies go here.
                    # Don't include "ReviewBoard" in this list.
                ],
                entry_points={
                    'reviewboard.extensions': [
                        '%(package_name)s = %(ext_class_path)s',
                    ],
                },
                classifiers=[
                    # For a full list of package classifiers, see
                    # %(classifiers_url)s

                    'Development Status :: 3 - Alpha',
                    'Environment :: Web Framework',
                    'Framework :: Review Board',
                    'Operating System :: OS Independent',
                    'Programming Language :: Python',
                ],
            )
        """ % {
            'author': self._sanitize_string_for_python(author or
                                                       '<REPLACE ME>'),
            'author_email': self._sanitize_string_for_python(author_email or
                                                             '<REPLACE ME>'),
            'classifiers_url':
                'https://pypi.python.org/pypi?%3Aaction=list_classifiers',
            'description': self._sanitize_string_for_python(summary or
                                                            '<REPLACE ME>'),
            'ext_class_path': '%s.extension:%s' % (package_name, class_name),
            'package_name': package_name,
            'version': version,
        }

    def _create_extension_py(self, name, package_name, class_name, summary,
                             configurable, has_static_media):
        """Create the content for an extension.py file.

        Args:
            name (unicode):
                The name of the extension.

            package_name (unicode):
                The name of the package.

            class_name (unicode):
                The name of the extension class.

            summary (unicode):
                A summary of the extension.

            configurable (bool):
                Whether the package is set to be configurable.

            has_static_media (bool):
                Whether the package is set to have static media files.

        Returns:
            unicode:
            The resulting content for the file.
        """
        extension_docs_url = '%sextending/extensions/' % get_manual_url()

        static_media_content = """
                # You can create a list of CSS bundles to compile and ship
                # with your extension. These can include both *.css and
                # *.less (http://lesscss.org/) files. See
                # %(static_docs_url)s
                css_bundles = {
                    'my-bundle-name': {
                        'source_filenames': [
                            'css/style.less',
                        ],
                        'apply_to': ['my-view-url-name'],
                    },
                }

                # JavaScript bundles are also supported. These support
                # standard *.js files and *.es6.js files (which allow for
                # writing and transpiling ES6 JavaScript).
                js_bundles = {
                    'my-bundle-name': {
                        'source_filenames': [
                            'js/script.es6.js',
                            'js/another-script.js',
                        ],
                    },
                }
        """ % {
            'static_docs_url': '%sstatic-files/' % extension_docs_url,
        }

        configuration_content = """
                # Default values for any configuration settings for your
                # extension.
                default_settings = {
                    'my_field_1': 'my default value',
                    'my_field_2': False,
                }

                # Set is_configurable and define an admin_urls.py to add
                # a standard configuration page for your extension.
                # See %(configure_docs_url)s
                is_configurable = True
        """ % {
            'configure_docs_url': '%sconfiguration/' % extension_docs_url,
        }

        return '''
            """%(name)s for Review Board."""

            from __future__ import unicode_literals

            from django.utils.translation import ugettext_lazy as _
            from reviewboard.extensions.base import Extension
            from reviewboard.extensions.hooks import TemplateHook


            class %(class_name)s(Extension):
                """Internal description for your extension here."""

                metadata = {
                    'Name': _(%(metadata_name)s),
                    'Summary': _(%(metadata_summary)s),
                }
            %(extra_class_content)s
                def initialize(self):
                    """Initialize the extension."""
                    # Set up any hooks your extension needs here. See
                    # %(hooks_docs_url)s
                    TemplateHook(self,
                                 'before-login-form',
                                 '%(package_name)s/before-login-form.html')
        ''' % {
            'class_name': class_name,
            'extra_class_content': ''.join(
                extra_content
                for extra_content, should_add in (
                    (configuration_content, configurable),
                    (static_media_content, has_static_media))
                if should_add
            ),
            'hooks_docs_url': '%s#python-extension-hooks' % extension_docs_url,
            'metadata_name': self._sanitize_string_for_python(name),
            'metadata_summary': self._sanitize_string_for_python(
                summary or 'REPLACE ME'),
            'name': name,
            'package_name': package_name,
        }

    def _create_admin_urls_py(self, package_name, class_name, form_class_name):
        """Create the content for an admin_urls.py file.

        Args:
            package_name (unicode):
                The name of the package.

            class_name (unicode):
                The name of the extension class.

            form_class_name (unicode):
                The name of the extension settings form class.

        Returns:
            unicode:
            The resulting content for the file.
        """
        return '''
            """Administration and configuration URLs for the extension."""

            from __future__ import unicode_literals

            from django.conf.urls import url
            from reviewboard.extensions.views import configure_extension

            from %(package_name)s.extension import %(class_name)s
            from %(package_name)s.forms import %(form_class_name)s


            urlpatterns = [
                url(r'^$',
                    configure_extension,
                    {
                        'ext_class': %(class_name)s,
                        'form_class': %(form_class_name)s,
                    }),
            ]
        ''' % {
            'class_name': class_name,
            'form_class_name': form_class_name,
            'package_name': package_name,
        }

    def _create_forms_py(self, form_class_name):
        """Create the content for a forms.py file.

        Args:
            form_class_name (unicode):
                The name of the extension settings form class.

        Returns:
            unicode:
            The resulting content for the file.
        """
        return '''
            """Configuration forms for the extension."""

            from __future__ import unicode_literals

            from django import forms
            from djblets.extensions.forms import SettingsForm


            class %(form_class_name)s(SettingsForm):
                my_field_1 = forms.CharField()
                my_field_2 = forms.BooleanField()
        ''' % {
            'form_class_name': form_class_name,
        }

    def _sanitize_string_for_python(self, s):
        """Sanitize a string for inclusion in a Python source file.

        This will return a string representation without any leading ``u``
        (when run on Python 2.x).

        Args:
            s (unicode):
                The string to sanitize.

        Returns:
            unicode:
            The sanitized string.
        """
        s = repr(s)

        if s.startswith('u'):
            s = s[1:]

        return s


class RBExt(object):
    """Command line tool for helping develop Review Board extensions.

    This tool provides subcommands useful for extension developers. It
    currently provides:

    * ``test``: Runs an extension's test suite.
    """

    COMMANDS = [
        CreateCommand(),
        TestCommand(),
    ]

    def run(self, argv):
        """Run an rbext command with the provided arguments.

        During the duration of the run, :py:data:`sys.argv` will be set to
        the provided arguments.

        Args:
            argv (list of unicode):
                The command line arguments passed to the command. This should
                not include the executable name as the first element.

        Returns:
            int:
            The command's exit code.
        """
        command, options = self.parse_options(argv)

        # We call out to things like the test runner, which expect to operate
        # off of sys.argv. We want to simulate that now that we've parsed
        # options. We'll restore sys.argv after the command finishes.
        old_argv = sys.argv
        sys.argv = argv

        try:
            return command.run(options)
        except Exception as e:
            logging.exception('Unexpected exception when running command '
                              '"%s": %s',
                              command.name, e)
            return 1
        finally:
            sys.argv = old_argv

    def parse_options(self, argv):
        """Parse arguments for the command.

        Args:
            argv (list of unicode):
                The arguments provided on the command line.

        Returns:
            unicode:
            The name of the command to run.
        """
        parser = argparse.ArgumentParser(prog='rbext',
                                         usage='%(prog)s <command>')

        subparsers = parser.add_subparsers(
            title='Commands',
            dest='command')

        commands = sorted(self.COMMANDS, key=lambda cmd: cmd.name)
        command_map = {}

        for command in commands:
            command_map[command.name] = command

            subparser = subparsers.add_parser(
                command.name,
                help=command.help_summary)

            subparser.add_argument(
                '-d',
                '--debug',
                action='store_true',
                dest='debug',
                default=False,
                help='Display debug output.')
            subparser.add_argument(
                '-s',
                '--settings-file',
                dest='settings_file',
                default=None,
                help='test_settings.py file to use for any custom settings.')

            command.add_options(subparser)

        # Prevent the '--' and anything after it from being parsed, so the
        # command can work with it.
        try:
            i = argv.index('--')
            argv = argv[:i]
        except ValueError:
            # The "--" isn't in the list anywhere.
            pass

        options = parser.parse_args(argv)

        return command_map[options.command], options


def main():
    """Run rbext.

    This is used by the Python EntryPoint to run rbext. It will pass in any
    arguments found on the command line and exit with the correct error code.
    """
    sys.exit(RBExt().run(sys.argv[1:]))



from __future__ import unicode_literals

from djblets.util.decorators import augment_method_from

from reviewboard.accounts.models import ReviewRequestVisit
from reviewboard.webapi.decorators import webapi_check_local_site
from reviewboard.webapi.resources.base_archived_object import \
    BaseArchivedObjectResource


class ArchivedReviewRequestResource(BaseArchivedObjectResource):
    """List and manipulate entries for review requests archived by the user.

    These are requests that the user has archived. This resource can be used
    for adding and removing archived review requests.

    Each item in the resource is an association between the user and the
    review request.
    """
    name = 'archived_review_request'
    visibility = ReviewRequestVisit.ARCHIVED

    @webapi_check_local_site
    @augment_method_from(BaseArchivedObjectResource)
    def create(self, *args, **kwargs):
        """Mark a review request as archived.

        The ID of the review reqiest must be passed as ``object_id``, and will
        store that review request in the list.
        """
        pass

    @webapi_check_local_site
    @augment_method_from(BaseArchivedObjectResource)
    def delete(self, *args, **kwargs):
        """Delete an archived review request entry.

        This is the same effect as unarchiving a review request. It does
        not actually delete the review request, just the entry in the list.
        """
        pass

archived_review_request_resource = ArchivedReviewRequestResource()



from __future__ import unicode_literals

from djblets.conditions import ConditionSet, Condition

from reviewboard.scmtools.conditions import (IsRepositoryPrivateOperator,
                                             IsRepositoryPublicOperator,
                                             RepositoryTypeChoice,
                                             RepositoriesChoice)
from reviewboard.site.models import LocalSite
from reviewboard.testing.testcase import TestCase


class RepositoryOperatorTests(TestCase):
    """Unit tests for repository condition operators."""

    fixtures = ['test_scmtools']

    def test_is_private_with_match(self):
        """Testing IsRepositoryPrivateOperator with match"""
        repository = self.create_repository(public=False)

        self.assertTrue(self._check_match(
            IsRepositoryPrivateOperator,
            repository))

    def test_is_private_without_match(self):
        """Testing IsRepositoryPrivateOperator without match"""
        repository = self.create_repository(public=True)

        self.assertFalse(self._check_match(
            IsRepositoryPrivateOperator,
            repository))
        self.assertFalse(self._check_match(
            IsRepositoryPrivateOperator,
            None))

    def test_is_public_with_match(self):
        """Testing IsRepositoryPublicOperator with match"""
        repository = self.create_repository(public=True)

        self.assertTrue(self._check_match(
            IsRepositoryPublicOperator,
            repository))

    def test_is_public_without_match(self):
        """Testing IsRepositoryPublicOperator without match"""
        repository = self.create_repository(public=False)

        self.assertFalse(self._check_match(
            IsRepositoryPublicOperator,
            repository))
        self.assertFalse(self._check_match(
            IsRepositoryPublicOperator,
            None))

    def _check_match(self, op_cls, match_value, condition_value=None):
        op = op_cls(None)
        return op.matches(match_value=match_value,
                          condition_value=condition_value)


class RepositoriesChoiceTests(TestCase):
    """Unit tests for reviewboard.scmtools.conditions.RepositoriesChoice."""

    fixtures = ['test_scmtools']

    def setUp(self):
        super(RepositoriesChoiceTests, self).setUp()

        self.choice = RepositoriesChoice()

    def test_get_queryset(self):
        """Testing RepositoriesChoice.get_queryset"""
        repo1 = self.create_repository(name='repo1')
        repo2 = self.create_repository(name='repo2')

        self.assertQuerysetEqual(
            self.choice.get_queryset().order_by('id'),
            [repo1.pk, repo2.pk],
            transform=lambda repo: repo.pk)

    def test_get_queryset_with_local_site(self):
        """Testing RepositoriesChoice.get_queryset with LocalSite"""
        good_site = LocalSite.objects.create(name='good-site')
        bad_site = LocalSite.objects.create(name='bad-site')

        # These should match.
        repo1 = self.create_repository(name='repo1', local_site=good_site)
        repo2 = self.create_repository(name='repo2', local_site=good_site)

        # These should not match.
        self.create_repository(name='repo3')
        self.create_repository(name='repo4', local_site=bad_site)

        self.choice.extra_state['local_site'] = good_site

        self.assertQuerysetEqual(
            self.choice.get_queryset().order_by('id'),
            [repo1.pk, repo2.pk],
            transform=lambda repo: repo.pk)

    def test_matches_with_any_op(self):
        """Testing RepositoriesChoice.matches with "any" operator"""
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('any')),
        ])

        self.assertTrue(condition_set.matches(
            repository=self.create_repository()))
        self.assertFalse(condition_set.matches(repository=None))

    def test_matches_with_none_op(self):
        """Testing RepositoriesChoice.matches with "none" operator"""
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('none')),
        ])

        self.assertTrue(condition_set.matches(repository=None))
        self.assertFalse(condition_set.matches(
            repository=self.create_repository()))

    def test_matches_with_one_of_op(self):
        """Testing RepositoriesChoice.matches with "one-of" operator"""
        repository1 = self.create_repository(name='repo1')
        repository2 = self.create_repository(name='repo2')
        repository3 = self.create_repository(name='repo3')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('one-of'),
                      [repository1, repository2])
        ])

        self.assertTrue(condition_set.matches(repository=repository1))
        self.assertTrue(condition_set.matches(repository=repository2))
        self.assertFalse(condition_set.matches(repository=repository3))

    def test_matches_with_not_one_of_op(self):
        """Testing RepositoriesChoice.matches with "not-one-of" operator"""
        repository1 = self.create_repository(name='repo1')
        repository2 = self.create_repository(name='repo2')
        repository3 = self.create_repository(name='repo3')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('not-one-of'),
                      [repository1, repository2])
        ])

        self.assertFalse(condition_set.matches(repository=repository1))
        self.assertFalse(condition_set.matches(repository=repository2))
        self.assertTrue(condition_set.matches(repository=repository3))

    def test_matches_with_is_public_op(self):
        """Testing RepositoriesChoice.matches with "is-public" operator"""
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('is-public')),
        ])

        self.assertTrue(condition_set.matches(
            repository=self.create_repository(name='repo1', public=True)))
        self.assertFalse(condition_set.matches(
            repository=self.create_repository(name='repo2', public=False)))

    def test_matches_with_is_private_op(self):
        """Testing RepositoriesChoice.matches with "is-private" operator"""
        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('is-private')),
        ])

        self.assertTrue(condition_set.matches(
            repository=self.create_repository(name='repo1', public=False)))
        self.assertFalse(condition_set.matches(
            repository=self.create_repository(name='repo2', public=True)))


class RepositoryTypeChoiceTests(TestCase):
    """Unit tests for reviewboard.scmtools.conditions.RepositoryTypeChoice."""

    fixtures = ['test_scmtools']

    def setUp(self):
        super(RepositoryTypeChoiceTests, self).setUp()

        self.choice = RepositoryTypeChoice()

    def test_matches_with_one_of_op(self):
        """Testing RepositoryTypeChoice.matches with "one-of" operator"""
        repository1 = self.create_repository(name='repo1',
                                             tool_name='Git')
        repository2 = self.create_repository(name='repo2',
                                             tool_name='Subversion')
        repository3 = self.create_repository(name='repo3',
                                             tool_name='CVS')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('one-of'),
                      [repository1.tool, repository2.tool])
        ])

        self.assertTrue(condition_set.matches(repository=repository1))
        self.assertTrue(condition_set.matches(repository=repository2))
        self.assertFalse(condition_set.matches(repository=repository3))

    def test_matches_with_not_one_of_op(self):
        """Testing RepositoryTypeChoice.matches with "not-one-of" operator"""
        repository1 = self.create_repository(name='repo1',
                                             tool_name='Git')
        repository2 = self.create_repository(name='repo2',
                                             tool_name='Subversion')
        repository3 = self.create_repository(name='repo3',
                                             tool_name='CVS')

        condition_set = ConditionSet(ConditionSet.MODE_ALL, [
            Condition(self.choice, self.choice.get_operator('not-one-of'),
                      [repository1.tool, repository2.tool])
        ])

        self.assertFalse(condition_set.matches(repository=repository1))
        self.assertFalse(condition_set.matches(repository=repository2))
        self.assertTrue(condition_set.matches(repository=repository3))



"""Middleware for Local Sites."""

from __future__ import unicode_literals

from django.utils.functional import SimpleLazyObject
from djblets.db.query import get_object_or_none

from reviewboard.site.models import LocalSite


class LocalSiteMiddleware(object):
    """Middleware that handles storing information on the Local Site in use.

    This adds a new ``local_site`` attribute to the
    :py:class:`~django.http.HttpRequest` that, when first accessed, will fetch
    and cache the matching :py:class:`~reviewboard.site.models.LocalSite`. If
    there's no Local Site for this given request, this will store ``None``
    instead.
    """

    def process_view(self, request, view_func, view_args, view_kwargs):
        """Process the request before calling the view.

        This sets up a ``local_site`` attribute on the request to fetch the
        :py:class:`~reviewboard.site.models.LocalSite` used for this view,
        if any. This is based on the ``local_site_name`` key in
        ``view_kwargs``.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            view_func (callable):
                The view being called. This is unused.

            view_args (tuple):
                The positional arguments passed to the view. This is unused.

            view_kwargs (dict):
                The keyword arguments passed to the view.
        """
        local_site_name = view_kwargs.get('local_site_name')
        request._local_site_name = local_site_name

        if request._local_site_name:
            request.local_site = SimpleLazyObject(
                lambda: get_object_or_none(LocalSite, name=local_site_name))
        else:
            request.local_site = None



from __future__ import unicode_literals

from django.test import RequestFactory
from djblets.forms.fields import ConditionsField

from reviewboard.integrations.base import Integration, get_integration_manager
from reviewboard.integrations.forms import IntegrationConfigForm
from reviewboard.reviews.conditions import ReviewRequestConditionChoices
from reviewboard.site.models import LocalSite
from reviewboard.testing.testcase import TestCase


class MyIntegration(Integration):
    pass


class MyConfigForm(IntegrationConfigForm):
    my_conditions = ConditionsField(ReviewRequestConditionChoices)


class IntegrationConfigFormTests(TestCase):
    """Unit tests for reviewboard.integrations.forms.IntegrationConfigForm."""

    def setUp(self):
        super(IntegrationConfigFormTests, self).setUp()

        self.integration = MyIntegration(get_integration_manager())
        self.request = RequestFactory().request()

    def test_init(self):
        """Testing IntegrationConfigForm initialization"""
        form = MyConfigForm(self.integration, self.request)

        local_site_1 = LocalSite.objects.create(name='local-site-1')
        local_site_2 = LocalSite.objects.create(name='local-site-2')

        local_site_ids = list(
            form.fields['local_site'].queryset
            .order_by('pk')
            .values_list('pk', flat=True)
        )

        self.assertIsNone(form.limit_to_local_site)
        self.assertEqual(local_site_ids, [local_site_1.pk, local_site_2.pk])
        self.assertEqual(form.fields['my_conditions'].choice_kwargs, {})

    def test_init_with_limit_to_local_site(self):
        """Testing IntegrationConfigForm initialization with
        limit_to_local_site
        """
        LocalSite.objects.create(name='local-site-1')
        local_site = LocalSite.objects.create(name='local-site-2')

        form = MyConfigForm(self.integration, self.request,
                            limit_to_local_site=local_site)

        local_site_ids = list(
            form.fields['local_site'].queryset
            .order_by('pk')
            .values_list('pk', flat=True)
        )

        self.assertEqual(form.limit_to_local_site, local_site)
        self.assertEqual(local_site_ids, [local_site.pk])
        self.assertEqual(
            form.fields['my_conditions'].choice_kwargs,
            {
                'local_site': local_site,
            })



from __future__ import unicode_literals

from django.conf.urls import include, url
from djblets.urls.resolvers import DynamicURLResolver


dynamic_urls = DynamicURLResolver()


urlpatterns = [
    url(r'^repos/(?P<repository_id>\d+)/', include([dynamic_urls])),
]



"""Unit tests for ActiveDirectoryBackend."""

from __future__ import unicode_literals

from django.contrib.auth.models import User

from reviewboard.accounts.backends import ActiveDirectoryBackend
from reviewboard.testing import TestCase


class ActiveDirectoryBackendTests(TestCase):
    """Unit tests for ActiveDirectoryBackend."""

    def test_get_or_create_user_without_ad_user_data_and_with_user(self):
        """Testing ActiveDirectoryBackend.get_or_create_user without
        ad_user_data and with user in database
        """
        backend = ActiveDirectoryBackend()
        user = User.objects.create(username='test')

        self.assertEqual(backend.get_or_create_user('test', None),
                         user)

    def test_get_or_create_user_without_ad_user_data_and_without_user(self):
        """Testing ActiveDirectoryBackend.get_or_create_user without
        ad_user_data and with user not in database
        """
        backend = ActiveDirectoryBackend()

        self.assertIsNone(backend.get_or_create_user('test', None))



from __future__ import unicode_literals

import bz2
import logging

from django.db import models
from django.db.models import Q
from django.utils import six, timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from djblets.db.fields import Base64Field, JSONField

from reviewboard.diffviewer.errors import DiffParserError
from reviewboard.diffviewer.managers import (RawFileDiffDataManager,
                                             FileDiffManager,
                                             DiffSetManager)
from reviewboard.scmtools.core import PRE_CREATION
from reviewboard.scmtools.models import Repository


class LegacyFileDiffData(models.Model):
    """Deprecated, legacy class for base64-encoded diff data.

    This is no longer populated, and exists solely to store legacy data
    that has not been migrated to :py:class:`RawFileDiffData`.
    """
    binary_hash = models.CharField(_("hash"), max_length=40, primary_key=True)
    binary = Base64Field(_("base64"))

    extra_data = JSONField(null=True)

    class Meta:
        db_table = 'diffviewer_filediffdata'
        verbose_name = _('Legacy File Diff Data')
        verbose_name_plural = _('Legacy File Diff Data Blobs')


class RawFileDiffData(models.Model):
    """Stores raw diff data as binary content in the database.

    This is the class used in Review Board 2.5+ to store diff content.
    Unlike in previous versions, the content is not base64-encoded. Instead,
    it is stored either as bzip2-compressed data (if the resulting
    compressed data is smaller than the raw data), or as the raw data itself.
    """
    COMPRESSION_BZIP2 = 'B'

    COMPRESSION_CHOICES = (
        (COMPRESSION_BZIP2, _('BZip2-compressed')),
    )

    binary_hash = models.CharField(_("hash"), max_length=40, unique=True)
    binary = models.BinaryField()
    compression = models.CharField(max_length=1, choices=COMPRESSION_CHOICES,
                                   null=True, blank=True)
    extra_data = JSONField(null=True)

    objects = RawFileDiffDataManager()

    @property
    def content(self):
        """Returns the content of the diff.

        The content will be uncompressed (if necessary) and returned as the
        raw set of bytes originally uploaded.
        """
        if self.compression == self.COMPRESSION_BZIP2:
            return bz2.decompress(self.binary)
        elif self.compression is None:
            return bytes(self.binary)
        else:
            raise NotImplementedError(
                'Unsupported compression method %s for RawFileDiffData %s'
                % (self.compression, self.pk))

    @property
    def insert_count(self):
        return self.extra_data.get('insert_count')

    @insert_count.setter
    def insert_count(self, value):
        self.extra_data['insert_count'] = value

    @property
    def delete_count(self):
        return self.extra_data.get('delete_count')

    @delete_count.setter
    def delete_count(self, value):
        self.extra_data['delete_count'] = value

    def recalculate_line_counts(self, tool):
        """Recalculates the insert_count and delete_count values.

        This will attempt to re-parse the stored diff and fetch the
        line counts through the parser.
        """
        logging.debug('Recalculating insert/delete line counts on '
                      'RawFileDiffData %s' % self.pk)

        try:
            files = tool.get_parser(self.content).parse()

            if len(files) != 1:
                raise DiffParserError(
                    'Got wrong number of files (%d)' % len(files))
        except DiffParserError as e:
            logging.error('Failed to correctly parse stored diff data in '
                          'RawFileDiffData ID %s when trying to get '
                          'insert/delete line counts: %s',
                          self.pk, e)
        else:
            file_info = files[0]
            self.insert_count = file_info.insert_count
            self.delete_count = file_info.delete_count

            if self.pk:
                self.save(update_fields=['extra_data'])

    class Meta:
        db_table = 'diffviewer_rawfilediffdata'
        verbose_name = _('Raw File Diff Data')
        verbose_name_plural = _('Raw File Diff Data Blobs')


@python_2_unicode_compatible
class FileDiff(models.Model):
    """
    A diff of a single file.

    This contains the patch and information needed to produce original and
    patched versions of a single file in a repository.
    """
    COPIED = 'C'
    DELETED = 'D'
    MODIFIED = 'M'
    MOVED = 'V'

    STATUSES = (
        (COPIED, _('Copied')),
        (DELETED, _('Deleted')),
        (MODIFIED, _('Modified')),
        (MOVED, _('Moved')),
    )

    _IS_PARENT_EMPTY_KEY = '__parent_diff_empty'

    diffset = models.ForeignKey('DiffSet',
                                related_name='files',
                                verbose_name=_("diff set"))

    source_file = models.CharField(_("source file"), max_length=1024)
    dest_file = models.CharField(_("destination file"), max_length=1024)
    source_revision = models.CharField(_("source file revision"),
                                       max_length=512)
    dest_detail = models.CharField(_("destination file details"),
                                   max_length=512)
    binary = models.BooleanField(_("binary file"), default=False)
    status = models.CharField(_("status"), max_length=1, choices=STATUSES)

    diff64 = Base64Field(
        _("diff"),
        db_column="diff_base64",
        blank=True)
    legacy_diff_hash = models.ForeignKey(
        LegacyFileDiffData,
        db_column='diff_hash_id',
        related_name='filediffs',
        null=True,
        blank=True)
    diff_hash = models.ForeignKey(
        RawFileDiffData,
        db_column='raw_diff_hash_id',
        related_name='filediffs',
        null=True,
        blank=True)

    parent_diff64 = Base64Field(
        _("parent diff"),
        db_column="parent_diff_base64",
        blank=True)
    legacy_parent_diff_hash = models.ForeignKey(
        LegacyFileDiffData,
        db_column='parent_diff_hash_id',
        related_name='parent_filediffs',
        null=True,
        blank=True)
    parent_diff_hash = models.ForeignKey(
        RawFileDiffData,
        db_column='raw_parent_diff_hash_id',
        related_name='parent_filediffs',
        null=True,
        blank=True)

    extra_data = JSONField(null=True)

    objects = FileDiffManager()

    @property
    def source_file_display(self):
        tool = self.diffset.repository.get_scmtool()
        return tool.normalize_path_for_display(self.source_file)

    @property
    def dest_file_display(self):
        tool = self.diffset.repository.get_scmtool()
        return tool.normalize_path_for_display(self.dest_file)

    @property
    def deleted(self):
        return self.status == self.DELETED

    @property
    def copied(self):
        return self.status == self.COPIED

    @property
    def moved(self):
        return self.status == self.MOVED

    @property
    def modified(self):
        """Whether this file is a modification to an existing file."""
        return self.status == self.MODIFIED

    @property
    def is_new(self):
        return self.source_revision == PRE_CREATION

    @property
    def status_string(self):
        """The FileDiff's status as a human-readable string."""
        if self.status == FileDiff.COPIED:
            return 'copied'
        elif self.status == FileDiff.DELETED:
            return 'deleted'
        elif self.status == FileDiff.MODIFIED:
            return 'modified'
        elif self.status == FileDiff.MOVED:
            return 'moved'
        else:
            logging.error('Unknown FileDiff status %r for FileDiff %s',
                          self.status, self.pk)
            return 'unknown'

    def _get_diff(self):
        if self._needs_diff_migration():
            self._migrate_diff_data()

        return self.diff_hash.content

    def _set_diff(self, diff):
        # Add hash to table if it doesn't exist, and set diff_hash to this.
        self.diff_hash, is_new = \
            RawFileDiffData.objects.get_or_create_from_data(diff)
        self.diff64 = ""

        return is_new

    diff = property(_get_diff, _set_diff)

    def _get_parent_diff(self):
        if self._needs_parent_diff_migration():
            self._migrate_diff_data()

        if self.parent_diff_hash:
            return self.parent_diff_hash.content
        else:
            return None

    def _set_parent_diff(self, parent_diff):
        if not parent_diff:
            return False

        # Add hash to table if it doesn't exist, and set diff_hash to this.
        self.parent_diff_hash, is_new = \
            RawFileDiffData.objects.get_or_create_from_data(parent_diff)
        self.parent_diff64 = ''

        return is_new

    parent_diff = property(_get_parent_diff, _set_parent_diff)

    def is_parent_diff_empty(self, cache_only=False):
        """Return whether or not the parent diff is empty.

        Args:
            cache_only (bool, optional):
                Whether or not to only use cached results.

        Returns:
            bool:
            Whether or not the parent diff is empty. This is true if either
            there is no parent diff or if the parent diff has no insertions and
            no deletions.
        """
        assert self.parent_diff_hash_id is not None, 'No parent diff.'

        if cache_only:
            return self.extra_data.get(self._IS_PARENT_EMPTY_KEY, False)

        if (not self.extra_data or
            self._IS_PARENT_EMPTY_KEY not in self.extra_data):
            parent_diff_hash = self.parent_diff_hash

            if (parent_diff_hash.insert_count is None or
                parent_diff_hash.delete_count is None):
                tool = self.diffset.repository.get_scmtool()
                parent_diff_hash.recalculate_line_counts(tool)

            self.extra_data[self._IS_PARENT_EMPTY_KEY] = (
                parent_diff_hash.insert_count == 0 and
                parent_diff_hash.delete_count == 0)

            self.save(update_fields=('extra_data',))

        return self.extra_data[self._IS_PARENT_EMPTY_KEY]

    @property
    def orig_sha1(self):
        return self.extra_data.get('orig_sha1')

    @property
    def patched_sha1(self):
        return self.extra_data.get('patched_sha1')

    def get_line_counts(self):
        """Returns the stored line counts for the diff.

        This will return all the types of line counts that can be set:

        * ``raw_insert_count``
        * ``raw_delete_count``
        * ``insert_count``
        * ``delete_count``
        * ``replace_count``
        * ``equal_count``
        * ``total_line_count``

        These are not all guaranteed to have values set, and may instead be
        None. Only ``raw_insert_count``, ``raw_delete_count``
        ``insert_count``, and ``delete_count`` are guaranteed to have values
        set.

        If there isn't a processed number of inserts or deletes stored,
        then ``insert_count`` and ``delete_count`` will be equal to the
        raw versions.
        """
        if ('raw_insert_count' not in self.extra_data or
            'raw_delete_count' not in self.extra_data):
            if not self.diff_hash:
                self._migrate_diff_data()

            if self.diff_hash.insert_count is None:
                self._recalculate_line_counts(self.diff_hash)

            self.extra_data.update({
                'raw_insert_count': self.diff_hash.insert_count,
                'raw_delete_count': self.diff_hash.delete_count,
            })

            if self.pk:
                self.save(update_fields=['extra_data'])

        raw_insert_count = self.extra_data['raw_insert_count']
        raw_delete_count = self.extra_data['raw_delete_count']

        return {
            'raw_insert_count': raw_insert_count,
            'raw_delete_count': raw_delete_count,
            'insert_count': self.extra_data.get('insert_count',
                                                raw_insert_count),
            'delete_count': self.extra_data.get('delete_count',
                                                raw_delete_count),
            'replace_count': self.extra_data.get('replace_count'),
            'equal_count': self.extra_data.get('equal_count'),
            'total_line_count': self.extra_data.get('total_line_count'),
        }

    def set_line_counts(self, raw_insert_count=None, raw_delete_count=None,
                        insert_count=None, delete_count=None,
                        replace_count=None, equal_count=None,
                        total_line_count=None):
        """Sets the line counts on the FileDiff.

        There are many types of useful line counts that can be set.

        ``raw_insert_count`` and ``raw_delete_count`` correspond to the
        raw inserts and deletes in the actual patch, which will be set both
        in this FileDiff and in the associated RawFileDiffData.

        The other counts are stored exclusively in FileDiff, as they are
        more render-specific.
        """
        updated = False

        if not self.diff_hash_id:
            # This really shouldn't happen, but if it does, we should handle
            # it gracefully.
            logging.warning('Attempting to call set_line_counts on '
                            'un-migrated FileDiff %s' % self.pk)
            self._migrate_diff_data(False)

        if (insert_count is not None and
            raw_insert_count is not None and
            self.diff_hash.insert_count is not None and
            self.diff_hash.insert_count != insert_count):
            # Allow overriding, but warn. This really shouldn't be called.
            logging.warning('Attempting to override insert count on '
                            'RawFileDiffData %s from %s to %s (FileDiff %s)'
                            % (self.diff_hash.pk,
                               self.diff_hash.insert_count,
                               insert_count,
                               self.pk))

        if (delete_count is not None and
            raw_delete_count is not None and
            self.diff_hash.delete_count is not None and
            self.diff_hash.delete_count != delete_count):
            # Allow overriding, but warn. This really shouldn't be called.
            logging.warning('Attempting to override delete count on '
                            'RawFileDiffData %s from %s to %s (FileDiff %s)'
                            % (self.diff_hash.pk,
                               self.diff_hash.delete_count,
                               delete_count,
                               self.pk))

        if raw_insert_count is not None or raw_delete_count is not None:
            # New raw counts have been provided. These apply to the actual
            # diff file itself, and will be common across all diffs sharing
            # the diff_hash instance. Set it there.
            if raw_insert_count is not None:
                self.diff_hash.insert_count = raw_insert_count
                self.extra_data['raw_insert_count'] = raw_insert_count
                updated = True

            if raw_delete_count is not None:
                self.diff_hash.delete_count = raw_delete_count
                self.extra_data['raw_delete_count'] = raw_delete_count
                updated = True

            self.diff_hash.save()

        for key, cur_value in (('insert_count', insert_count),
                               ('delete_count', delete_count),
                               ('replace_count', replace_count),
                               ('equal_count', equal_count),
                               ('total_line_count', total_line_count)):
            if cur_value is not None and cur_value != self.extra_data.get(key):
                self.extra_data[key] = cur_value
                updated = True

        if updated and self.pk:
            self.save(update_fields=['extra_data'])

    def _needs_diff_migration(self):
        return self.diff_hash_id is None

    def _needs_parent_diff_migration(self):
        return (self.parent_diff_hash_id is None and
                (self.parent_diff64 or self.legacy_parent_diff_hash_id))

    def _migrate_diff_data(self, recalculate_counts=True):
        """Migrates diff data associated with a FileDiff to RawFileDiffData.

        If the diff data is stored directly on the FileDiff, it will be
        removed and stored on a RawFileDiffData instead.

        If the diff data is stored on an associated LegacyFileDiffData,
        that will be converted into a RawFileDiffData. The LegacyFileDiffData
        will then be removed, if nothing else is using it.
        """
        needs_save = False
        diff_hash_is_new = False
        parent_diff_hash_is_new = False
        fix_refs = False
        legacy_pks = []
        needs_diff_migration = self._needs_diff_migration()
        needs_parent_diff_migration = self._needs_parent_diff_migration()

        if needs_diff_migration:
            recalculate_diff_counts = recalculate_counts
            needs_save = True

            if self.legacy_diff_hash_id:
                logging.debug('Migrating LegacyFileDiffData %s to '
                              'RawFileDiffData for diff in FileDiff %s',
                              self.legacy_diff_hash_id, self.pk)

                try:
                    legacy_data = self.legacy_diff_hash.binary
                except LegacyFileDiffData.DoesNotExist:
                    # Another process migrated this before we could.
                    # We'll need to fix the references.
                    fix_refs = True
                    recalculate_diff_counts = False
                else:
                    diff_hash_is_new = self._set_diff(legacy_data)
                    legacy_pks.append(self.legacy_diff_hash_id)
                    self.legacy_diff_hash = None
            else:
                logging.debug('Migrating FileDiff %s diff data to '
                              'RawFileDiffData',
                              self.pk)

                diff_hash_is_new = self._set_diff(self.diff64)

            if recalculate_diff_counts:
                self._recalculate_line_counts(self.diff_hash)

        if needs_parent_diff_migration:
            recalculate_parent_diff_counts = recalculate_counts
            needs_save = True

            if self.legacy_parent_diff_hash_id:
                logging.debug('Migrating LegacyFileDiffData %s to '
                              'RawFileDiffData for parent diff in FileDiff %s',
                              self.legacy_parent_diff_hash_id, self.pk)

                try:
                    legacy_parent_data = self.legacy_parent_diff_hash.binary
                except LegacyFileDiffData.DoesNotExist:
                    # Another process migrated this before we could.
                    # We'll need to fix the references.
                    fix_refs = True
                    recalculate_parent_diff_counts = False
                else:
                    parent_diff_hash_is_new = \
                        self._set_parent_diff(legacy_parent_data)
                    legacy_pks.append(self.legacy_parent_diff_hash_id)
                    self.legacy_parent_diff_hash = None
            else:
                logging.debug('Migrating FileDiff %s parent diff data to '
                              'RawFileDiffData',
                              self.pk)

                parent_diff_hash_is_new = \
                    self._set_parent_diff(self.parent_diff64)

            if recalculate_parent_diff_counts:
                self._recalculate_line_counts(self.parent_diff_hash)

        if fix_refs:
            # Another server/process/thread got to this before we could.
            # We need to pull the latest refs and make sure they're set here.
            diff_hash, parent_diff_hash = (
                FileDiff.objects.filter(pk=self.pk)
                .values_list('diff_hash_id', 'parent_diff_hash_id')[0]
            )

            if needs_diff_migration:
                if diff_hash:
                    self.diff_hash_id = diff_hash
                    self.legacy_diff_hash = None
                    self.diff64 = ''
                else:
                    logging.error('Unable to migrate diff for FileDiff %s: '
                                  'LegacyFileDiffData "%s" is missing, and '
                                  'database entry does not have a new '
                                  'diff_hash! Data may be missing.',
                                  self.pk, self.legacy_diff_hash_id)

            if needs_parent_diff_migration:
                if parent_diff_hash:
                    self.parent_diff_hash_id = parent_diff_hash
                    self.legacy_parent_diff_hash = None
                    self.parent_diff64 = ''
                else:
                    logging.error('Unable to migrate parent diff for '
                                  'FileDiff %s: LegacyFileDiffData "%s" is '
                                  'missing, and database entry does not have '
                                  'a new parent_diff_hash! Data may be '
                                  'missing.',
                                  self.pk, self.legacy_parent_diff_hash_id)

        if needs_save:
            if self.pk:
                self.save(update_fields=[
                    'diff64', 'parent_diff64', 'diff_hash', 'parent_diff_hash',
                    'legacy_diff_hash', 'legacy_parent_diff_hash',
                ])
            else:
                self.save()

        if legacy_pks:
            # Delete any LegacyFileDiffData objects no longer associated
            # with any FileDiffs.
            LegacyFileDiffData.objects \
                .filter(pk__in=legacy_pks) \
                .exclude(Q(filediffs__pk__gt=0) |
                         Q(parent_filediffs__pk__gt=0)) \
                .delete()

        return diff_hash_is_new, parent_diff_hash_is_new

    def _recalculate_line_counts(self, diff_hash):
        """Recalculates the line counts on the specified RawFileDiffData.

        This requires that diff_hash is set. Otherwise, it will assert.
        """
        diff_hash.recalculate_line_counts(
            self.diffset.repository.get_scmtool())

    def __str__(self):
        return "%s (%s) -> %s (%s)" % (self.source_file, self.source_revision,
                                       self.dest_file, self.dest_detail)

    class Meta:
        db_table = 'diffviewer_filediff'
        verbose_name = _('File Diff')
        verbose_name_plural = _('File Diffs')


@python_2_unicode_compatible
class DiffSet(models.Model):
    """
    A revisioned collection of FileDiffs.
    """
    name = models.CharField(_('name'), max_length=256)
    revision = models.IntegerField(_("revision"))
    timestamp = models.DateTimeField(_("timestamp"), default=timezone.now)
    basedir = models.CharField(_('base directory'), max_length=256,
                               blank=True, default='')
    history = models.ForeignKey('DiffSetHistory', null=True,
                                related_name="diffsets",
                                verbose_name=_("diff set history"))
    repository = models.ForeignKey(Repository, related_name="diffsets",
                                   verbose_name=_("repository"))
    diffcompat = models.IntegerField(
        _('differ compatibility version'),
        default=0,
        help_text=_("The diff generator compatibility version to use. "
                    "This can and should be ignored."))

    base_commit_id = models.CharField(
        _('commit ID'), max_length=64, blank=True, null=True, db_index=True,
        help_text=_('The ID/revision this change is built upon.'))

    extra_data = JSONField(null=True)

    objects = DiffSetManager()

    def get_total_line_counts(self):
        """Returns the total line counts from all files in this diffset."""
        counts = {}

        for filediff in self.files.all():
            for key, value in six.iteritems(filediff.get_line_counts()):
                if counts.get(key) is None:
                    counts[key] = value
                elif value is not None:
                    counts[key] += value

        return counts

    def update_revision_from_history(self, diffset_history):
        """Update the revision of this diffset based on a diffset history.

        This will determine the appropriate revision to use for the diffset,
        based on how many other diffsets there are in the history. If there
        aren't any, the revision will be set to 1.

        Args:
            diffset_history (reviewboard.diffviewer.models.DiffSetHistory):
                The diffset history used to compute the new revision.

        Raises:
            ValueError:
                The revision already has a valid value set, and cannot be
                updated.
        """
        if self.revision not in (0, None):
            raise ValueError('The diffset already has a valid revision set.')

        # Default this to revision 1. We'll use this if the DiffSetHistory
        # isn't saved yet (which may happen when creating a new review request)
        # or if there aren't yet any diffsets.
        self.revision = 1

        if diffset_history.pk:
            try:
                latest_diffset = \
                    diffset_history.diffsets.only('revision').latest()
                self.revision = latest_diffset.revision + 1
            except DiffSet.DoesNotExist:
                # Stay at revision 1.
                pass

    def save(self, **kwargs):
        """Save this diffset.

        This will set an initial revision of 1 if this is the first diffset
        in the history, and will set it to on more than the most recent
        diffset otherwise.

        Args:
            **kwargs (dict):
                Extra arguments for the save call.
        """
        if self.history is not None:
            if self.revision == 0:
                self.update_revision_from_history(self.history)

            self.history.last_diff_updated = self.timestamp
            self.history.save()

        super(DiffSet, self).save(**kwargs)

    def __str__(self):
        return "[%s] %s r%s" % (self.id, self.name, self.revision)

    class Meta:
        db_table = 'diffviewer_diffset'
        get_latest_by = 'revision'
        ordering = ['revision', 'timestamp']
        verbose_name = _('Diff Set')
        verbose_name_plural = _('Diff Sets')


@python_2_unicode_compatible
class DiffSetHistory(models.Model):
    """
    A collection of diffsets.

    This gives us a way to store and keep track of multiple revisions of
    diffsets belonging to an object.
    """
    name = models.CharField(_('name'), max_length=256)
    timestamp = models.DateTimeField(_("timestamp"), default=timezone.now)
    last_diff_updated = models.DateTimeField(
        _("last updated"),
        blank=True,
        null=True,
        default=None)

    extra_data = JSONField(null=True)

    def __str__(self):
        return 'Diff Set History (%s revisions)' % self.diffsets.count()

    class Meta:
        db_table = 'diffviewer_diffsethistory'
        verbose_name = _('Diff Set History')
        verbose_name_plural = _('Diff Set Histories')



"""Views used to render e-mail previews."""

from __future__ import unicode_literals

from django.views.generic.base import View

from reviewboard.notifications.email.decorators import preview_email


class BasePreviewEmailView(View):
    """Generic view used to preview rendered e-mails.

    This is a convenience for class-based views that wraps the
    :py:class:`@preview_email
    <reviewboard.notifications.email.decorators.preview_email>` decorator.
    """

    #: The function that will build the e-mail.
    #:
    #: This must be a class method or, if using an existing top-level function,
    #: :js:class:`staticmethod` must be used.
    build_email = None

    def get(self, *args, **kwargs):
        """Handle a HTTP GET request.

        The request will render the e-mail.

        Args:
            *args (tuple):
                Positional arguments passed to the handler.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response from the handler.
        """
        return preview_email(self.build_email)(self.get_email_data)(*args,
                                                                    **kwargs)

    def get_email_data(self, *args, **kwargs):
        """Return data used for the e-mail builder.

        The data returned will be passed to :py:attr:`build_email` to handle
        rendering the e-mail.

        This can also return a :py:class:`~django.http.HttpResponse`, which
        is useful for returning errors.

        Args:
            *args (tuple):
                Positional arguments passed to the handler.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            object:
            The dictionary data to pass as keyword arguments to
            :py:attr:`build_email`, or an instance of
            :py:class:`~django.http.HttpResponse` to immediately return to
            the client.
        """
        return {}



"""Unit tests for the Gitorious hosting service."""

from __future__ import unicode_literals

from reviewboard.hostingsvcs.tests.testcases import ServiceTests


class GitoriousTests(ServiceTests):
    """Unit tests for the Gitorious hosting service."""

    service_name = 'gitorious'

    def test_service_support(self):
        """Testing Gitorious service support capabilities"""
        self.assertFalse(self.service_class.supports_bug_trackers)
        self.assertTrue(self.service_class.supports_repositories)
        self.assertTrue(self.service_class.self_hosted)

    def test_get_repository_fields(self):
        """Testing Gitorious.get_repository_fields"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                fields={
                    'gitorious_project_name': 'myproj',
                    'gitorious_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'git://example.com/myproj/myrepo.git',
                'mirror_path': 'https://example.com/myproj/myrepo.git',
                'raw_file_url': ('https://example.com/myproj/myrepo/blobs/raw/'
                                 '<revision>'),
            })



"""Custom search index fields."""

from __future__ import unicode_literals

from haystack import indexes


class BooleanField(indexes.BooleanField):
    """A custom BooleanField.

    This works around `an issue in django-haystack
    <https://github.com/django-haystack/django-haystack/issues/801>`_ that
    results in BooleanFields always returning ``True``.
    """

    value_map = {
        'true': True,
        'false': False,
    }

    def convert(self, value):
        """Convert value to a boolean value.

        Args:
            value (unicode):
                The value to convert

        Returns:
            bool:
            The boolean representation of ``value``.
        """
        if value is None:
            return None

        try:
            return self.value_map[value]
        except KeyError:
            return bool(value)



from __future__ import print_function, unicode_literals

import os

from django.conf import settings
from django.contrib.auth.models import User
from django.core import mail
from django.utils import six
from djblets.siteconfig.models import SiteConfiguration
from djblets.webapi.testing.testcases import WebAPITestCaseMixin

from reviewboard.notifications.tests.test_email_sending import EmailTestHelper
from reviewboard.reviews.models import Review
from reviewboard.testing import TestCase
from reviewboard.webapi.tests.mimetypes import (
    error_mimetype,
    file_attachment_comment_item_mimetype,
    general_comment_item_mimetype,
    review_diff_comment_item_mimetype,
    screenshot_comment_item_mimetype)
from reviewboard.webapi.tests.urls import (
    get_review_diff_comment_list_url,
    get_review_file_attachment_comment_list_url,
    get_review_general_comment_list_url,
    get_review_screenshot_comment_list_url,
    get_screenshot_list_url)


class BaseWebAPITestCase(WebAPITestCaseMixin, TestCase, EmailTestHelper):
    error_mimetype = error_mimetype

    def setUp(self):
        super(BaseWebAPITestCase, self).setUp()

        self.siteconfig = SiteConfiguration.objects.get_current()
        self.siteconfig.set("mail_send_review_mail", False)
        self.siteconfig.set("auth_require_sitewide_login", False)
        self.siteconfig.save()
        self._saved_siteconfig_settings = self.siteconfig.settings.copy()

        mail.outbox = []

        fixtures = getattr(self, 'fixtures', None)

        if fixtures and 'test_users' in fixtures:
            self.client.login(username="grumpy", password="grumpy")
            self.user = User.objects.get(username="grumpy")

        self.base_url = 'http://testserver'

    def tearDown(self):
        super(BaseWebAPITestCase, self).tearDown()

        self.client.logout()

        if self.siteconfig.settings != self._saved_siteconfig_settings:
            self.siteconfig.settings = self._saved_siteconfig_settings
            self.siteconfig.save()

    def _testHttpCaching(self, url, check_etags=False,
                         check_last_modified=False):
        response = self.client.get(url)
        self.assertHttpOK(response, check_etag=check_etags,
                          check_last_modified=check_last_modified)

        headers = {}

        if check_etags:
            headers['HTTP_IF_NONE_MATCH'] = response['ETag']

        if check_last_modified:
            headers['HTTP_IF_MODIFIED_SINCE'] = response['Last-Modified']

        response = self.client.get(url, **headers)

        self.assertHttpNotModified(response)

    #
    # Some utility functions shared across test suites.
    #
    def _login_user(self, local_site=False, admin=False):
        """Creates a user for a test.

        The proper user will be created based on whether a valid LocalSite
        user is needed, and/or an admin user is needed.
        """
        self.client.logout()

        # doc is a member of the default LocalSite.
        username = 'doc'

        if admin:
            if local_site:
                user = User.objects.get(username=username)
                local_site = self.get_local_site(name=self.local_site_name)
                local_site.admins.add(user)
            else:
                username = 'admin'
        elif not local_site:
            # Pick a user that's not part of the default LocalSite.
            username = 'grumpy'

        self.assertTrue(self.client.login(username=username,
                                          password=username))

        return User.objects.get(username=username)

    def _postNewDiffComment(self, review_request, review_id, comment_text,
                            filediff_id=None, interfilediff_id=None,
                            first_line=10, num_lines=5, issue_opened=None,
                            issue_status=None):
        """Creates a diff comment and returns the payload response."""
        if filediff_id is None:
            diffset = review_request.diffset_history.diffsets.latest()
            filediff = diffset.files.all()[0]
            filediff_id = filediff.id

        data = {
            'filediff_id': filediff_id,
            'text': comment_text,
            'first_line': first_line,
            'num_lines': num_lines,
        }

        if interfilediff_id is not None:
            data['interfilediff_id'] = interfilediff_id

        if issue_opened is not None:
            data['issue_opened'] = issue_opened

        if issue_status is not None:
            data['issue_status'] = issue_status

        if review_request.local_site:
            local_site_name = review_request.local_site.name
        else:
            local_site_name = None

        review = Review.objects.get(pk=review_id)

        rsp = self.api_post(
            get_review_diff_comment_list_url(review, local_site_name),
            data,
            expected_mimetype=review_diff_comment_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')

        return rsp

    def _postNewScreenshotComment(self, review_request, review_id, screenshot,
                                  comment_text, x, y, w, h, issue_opened=None,
                                  issue_status=None):
        """Creates a screenshot comment and returns the payload response."""
        if review_request.local_site:
            local_site_name = review_request.local_site.name
        else:
            local_site_name = None

        post_data = {
            'screenshot_id': screenshot.id,
            'text': comment_text,
            'x': x,
            'y': y,
            'w': w,
            'h': h,
        }

        if issue_opened is not None:
            post_data['issue_opened'] = issue_opened

        if issue_status is not None:
            post_data['issue_status'] = issue_status

        review = Review.objects.get(pk=review_id)
        rsp = self.api_post(
            get_review_screenshot_comment_list_url(review, local_site_name),
            post_data,
            expected_mimetype=screenshot_comment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        return rsp

    def _delete_screenshot(self, review_request, screenshot):
        """Deletes a screenshot.

        This does not return anything, because DELETE requests don't return a
        response with a payload.
        """
        if review_request.local_site:
            local_site_name = review_request.local_site.name
        else:
            local_site_name = None

        self.api_delete(
            get_screenshot_list_url(review_request, local_site_name) +
            six.text_type(screenshot.id) + '/')

    def _postNewFileAttachmentComment(self, review_request, review_id,
                                      file_attachment, comment_text,
                                      issue_opened=None,
                                      issue_status=None,
                                      extra_fields={}):
        """Creates a file attachment comment.

        This returns the response from the API call to create the comment."""
        if review_request.local_site:
            local_site_name = review_request.local_site.name
        else:
            local_site_name = None

        post_data = {
            'file_attachment_id': file_attachment.id,
            'text': comment_text,
        }
        post_data.update(extra_fields)

        if issue_opened is not None:
            post_data['issue_opened'] = issue_opened

        if issue_status is not None:
            post_data['issue_status'] = issue_status

        review = Review.objects.get(pk=review_id)
        rsp = self.api_post(
            get_review_file_attachment_comment_list_url(review,
                                                        local_site_name),
            post_data,
            expected_mimetype=file_attachment_comment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        return rsp

    def _post_new_general_comment(self, review_request, review_id,
                                  comment_text,
                                  issue_opened=None,
                                  issue_status=None):
        """Creates a general comment.

        This returns the response from the API call to create the comment.
        """
        if review_request.local_site:
            local_site_name = review_request.local_site.name
        else:
            local_site_name = None

        post_data = {
            'text': comment_text,
        }

        if issue_opened is not None:
            post_data['issue_opened'] = issue_opened

        if issue_status is not None:
            post_data['issue_status'] = issue_status

        review = Review.objects.get(pk=review_id)
        rsp = self.api_post(
            get_review_general_comment_list_url(review, local_site_name),
            post_data,
            expected_mimetype=general_comment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        return rsp

    def get_sample_image_filename(self):
        return os.path.join(settings.STATIC_ROOT, 'rb', 'images', 'logo.png')



"""An API for managing OAuth2 tokens."""

from __future__ import unicode_literals

from django.db.models.query import Q
from django.utils import six
from django.utils.translation import ugettext_lazy as _
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_request_fields,
                                       webapi_response_errors)
from djblets.webapi.oauth2_scopes import get_scope_dictionary
from djblets.webapi.errors import DOES_NOT_EXIST, INVALID_FORM_DATA
from oauth2_provider.models import AccessToken

from reviewboard.oauth.features import oauth2_service_feature
from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.decorators import webapi_check_local_site


class OAuthTokenResource(WebAPIResource):
    """An API resource for managing OAuth2 tokens.

    This resource allows callers to list, update, or delete their existing
    tokens.
    """

    model = AccessToken
    name = 'oauth_token'
    verbose_name = _('OAuth2 Tokens')
    uri_object_key = 'oauth_token_id'
    item_result_key = 'oauth_token'
    required_features = [oauth2_service_feature]
    allowed_methods = ('GET', 'PUT', 'DELETE')

    api_token_access_allowed = False
    oauth2_token_access_allowed = False

    added_in = '3.0'

    fields = {
        'application': {
            'type': six.text_type,
            'description': 'The name of the application this token is for.',
        },
        'expires': {
            'type': six.text_type,
            'description': 'When this token is set to expire.',
        },
        'scope': {
            'type': [six.text_type],
            'description': 'The scopes this token has access to.',
        },
        'token': {
            'type': six.text_type,
            'description': 'The access token.',
        },
    }

    def serialize_application_field(self, obj, *args, **kwargs):
        """Serialize the application field.

        Args:
            obj (oauth2_provider.models.AccessToken):
                The token that is being serialized.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Ignored keyword arguments.

        Returns:
            unicode:
            The name of the application the access token has access to.
        """
        return obj.application.name

    def serialize_expires_field(self, obj, *args, **kwargs):
        """Serialize the expires field.

        Args:
            obj (oauth2_provider.models.AccessToken):
                The token that is being serialized.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Ignored keyword arguments.

        Returns:
            unicode:
            The expiry date of the token, in ISO-8601 format.
        """
        return obj.expires.isoformat()

    def serialize_scope_field(self, obj, *args, **kwargs):
        """Serialize the scope field.

        Args:
            obj (oauth2_provider.models.AccessToken):
                The token that is being serialized.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Ignored keyword arguments.

        Returns:
            list of unicode:
            The list of scopes the token has.
        """
        return obj.scope.split()

    def get_queryset(self, request, *args, **kwargs):
        """Return the queryset for the request.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            local_site (reviewboard.site.models.LocalSite, optional):
                The current LocalSite, if any.

        Returns:
            django.db.models.query.QuerySet:
            The tokens the user has access to.
        """
        if not request.user.is_authenticated():
            return AccessToken.objects.none()

        q = Q(application__local_site=request.local_site)

        if not request.user.is_superuser:
            q &= Q(user=request.user)

        return (
            AccessToken.objects
            .filter(q)
            .select_related('application')
        )

    def has_access_permissions(self, request, obj, *args, **kwargs):
        """Return whether or not the user has access permissions.

        A user has this permission if they own the token or are a superuser.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            obj (oauth2_provider.models.AccessToken):
                The token in question.

        Returns:
            bool:
            Whether or not the user has permission.
        """
        return (request.user.is_authenticated() and
                (obj.user == request.user or
                 request.user.is_superuser))

    def has_modify_permissions(self, request, obj, *args, **kwargs):
        """Return whether or not the user has modification permissions.

        A user has this permission if they own the token or are a superuser.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            obj (oauth2_provider.models.AccessToken):
                The token in question.

        Returns:
            bool:
            Whether or not the user has permission.
        """
        return self.has_access_permissions(request, obj, *args, **kwargs)

    def has_delete_permissions(self, request, obj, *args, **kwargs):
        """Return whether or not the user has deletion permissions.

        A user has this permission if they own the token or are a superuser.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            obj (oauth2_provider.models.AccessToken):
                The token in question.

        Returns:
            bool:
            Whether or not the user has permission.
        """
        return self.has_access_permissions(request, obj, *args, **kwargs)

    @webapi_login_required
    @augment_method_from(WebAPIResource)
    def get(self, *args, **kwargs):
        """Retrieves information on a particular OAuth2 token.

        This can only be accessed by the owner of the tokens or superusers
        """
        pass

    @webapi_login_required
    @augment_method_from(WebAPIResource)
    def get_list(self, *args, **kwargs):
        """Retrieve a list of information about an OAuth2 token.

        If accessing this API on a Local Site, the results will be limited
        to those associated with that site. Otherwise, it will be limited to
        those associated with no Local Site.

        This can only be accessed by the owner of the tokens or superusers.
        """
        pass

    @augment_method_from(WebAPIResource)
    def delete(self, *args, **kwargs):
        """Delete the OAuth2 token, invalidating all clients using it.

        The OAuth token will be removed from the user's account, and will no
        longer be usable for authentication.

        After deletion, this will return a :http:`204`.
        """
        pass

    @webapi_login_required
    @webapi_check_local_site
    @webapi_response_errors(DOES_NOT_EXIST)
    @webapi_request_fields(
        optional={
            'add_scopes': {
                'type': six.text_type,
                'description': 'A comma-separated list of scopes to add.',
            },
            'remove_scopes': {
                'type': six.text_type,
                'description': 'A comma-separated list of scopes to remove.',
            },
            'scopes': {
                'type': six.text_type,
                'description': 'A comma-separated list of scopes to override '
                               'the current set with.\n\n'
                               'This field cannot be provided if either '
                               'add_scopes or remove_scopes is provided.',
            },
        },
    )
    def update(self, request, local_site=None, add_scopes=None,
               remove_scopes=None, scopes=None, *args, **kwargs):
        """Update the scope of an OAuth2 token.

        This resource allows a user to either (1) add and remove scopes or (2)
        replace the set of scopes with a new set.
        """
        try:
            access_token = self.get_object(request, *args, **kwargs)
        except AccessToken.DoesNotExist:
            return DOES_NOT_EXIST

        if not self.has_modify_permissions(request, access_token, *args,
                                           **kwargs):
            return self.get_no_access_error(request)

        if ((add_scopes is not None or remove_scopes is not None) and
            scopes is not None):
            return INVALID_FORM_DATA, {
                'fields': {
                    'scopes': [
                        'This field cannot be provided if either add_scopes '
                        'or remove_scopes is provided.',
                    ],
                },
            }

        field_errors = {}
        valid_scopes = get_scope_dictionary()

        if scopes is not None:
            scopes = self._validate_scopes(valid_scopes, scopes, 'scopes',
                                           field_errors)
        elif add_scopes is not None or remove_scopes is not None:
            add_scopes = self._validate_scopes(valid_scopes,
                                               add_scopes,
                                               'add_scopes',
                                               field_errors)
            remove_scopes = self._validate_scopes(valid_scopes,
                                                  remove_scopes,
                                                  'remove_scopes',
                                                  field_errors)

        if field_errors:
            return INVALID_FORM_DATA, {
                'fields': field_errors,
            }

        if scopes is not None:
            access_token.scope = ' '.join(scopes)
            access_token.save(update_fields=('scope',))
        elif add_scopes is not None or remove_scopes is not None:
            current_scopes = set(access_token.scope.split(' '))

            if add_scopes:
                current_scopes.update(add_scopes)

            if remove_scopes:
                current_scopes.difference_update(remove_scopes)

            access_token.scope = ' '.join(current_scopes)
            access_token.save(update_fields=('scope',))

        return 200, {
            self.item_result_key: access_token,
        }

    def _validate_scopes(self, valid_scopes, scopes, field, field_errors):
        """Validate the given set of scopes against known valid scopes.

        Args:
            valid_scopes (dict):
                The scope dictionary.

            scopes (unicode):
                The comma-separated list of scopes to validate.

            field (unicode):
                The name of the field that is being validated.

            field_errors (dict):
                A mapping of field names to errors.

                An error message will be added to ``field_errors[field]`` for
                each invalid scope.

        Returns:
            list:
            The list of scopes, if they are all valid, or ``None`` otherwise.
        """
        if scopes is None:
            return None

        scopes = scopes.split(',')
        invalid_scopes = {
            scope
            for scope in scopes
            if scope not in valid_scopes
        }

        if invalid_scopes:
            field_errors[field] = [
                'The scope "%s" is invalid.' % scope
                for scope in invalid_scopes
            ]

            return None

        return scopes


oauth_token_resource = OAuthTokenResource()



"""Unit tests for reviewboard.accounts.models.Profile."""

from __future__ import unicode_literals

from django.contrib.auth.models import AnonymousUser, User
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.models import LocalSiteProfile
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class ProfileTests(TestCase):
    """Unit tests for reviewboard.accounts.models.Profile."""

    fixtures = ['test_users']

    @add_fixtures(['test_scmtools', 'test_site'])
    def test_is_star_unstar_updating_count_correctly(self):
        """Testing if star, unstar affect review request counts correctly"""
        user1 = User.objects.get(username='admin')
        profile1 = user1.get_profile()
        review_request = self.create_review_request(publish=True)

        site_profile = profile1.site_profiles.get(local_site=None)

        profile1.star_review_request(review_request)
        site_profile = LocalSiteProfile.objects.get(pk=site_profile.pk)

        self.assertTrue(review_request in
                        profile1.starred_review_requests.all())
        self.assertEqual(site_profile.starred_public_request_count, 1)

        profile1.unstar_review_request(review_request)
        site_profile = LocalSiteProfile.objects.get(pk=site_profile.pk)

        self.assertFalse(review_request in
                         profile1.starred_review_requests.all())
        self.assertEqual(site_profile.starred_public_request_count, 0)

    def test_get_display_name_unauthenticated_public(self):
        """Testing Profile.get_display_name with a public profile"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = False

        self.assertEqual(profile.get_display_name(AnonymousUser()),
                         user.username)

    def test_get_display_name_unauthenticated_private(self):
        """Testing Profile.get_display_name for an unauthenticated user viewing
        a private profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        self.assertEqual(profile.get_display_name(AnonymousUser()),
                         user.username)

    def test_get_display_name_public(self):
        """Testing Profile.get_display_name for an authenticated user viewing a
        public profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = False

        self.assertEqual(
            profile.get_display_name(User.objects.get(username='grumpy')),
            user.get_full_name())

    def test_get_display_name_private(self):
        """Testing Profile.get_display_name for an authenticated user viewing a
        private profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        self.assertEqual(
            profile.get_display_name(User.objects.get(username='grumpy')),
            user.username)

    def test_get_display_name_admin_private(self):
        """Testing Profile.get_display_name for an admin viewing a private
        profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        self.assertEqual(
            profile.get_display_name(User.objects.get(username='admin')),
            user.get_full_name())

    @add_fixtures(['test_site'])
    def test_get_display_name_localsite_member_private(self):
        """Testing Profile.get_display_name for a LocalSite member viewing
        a LocalSite member with a private profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        viewer = User.objects.get(username='grumpy')
        site = LocalSite.objects.get(pk=1)
        site.users.add(viewer)

        self.assertEqual(profile.get_display_name(viewer), user.username)

    @add_fixtures(['test_site'])
    def test_get_display_name_localsite_admin_private(self):
        """Testing Profile.get_display_name for a LocalSite admin viewing
        a LocalSite member with a private profile
        """
        user = User.objects.get(username='admin')
        profile = user.get_profile()
        profile.is_private = True

        self.assertEqual(
            profile.get_display_name(User.objects.get(username='doc')),
            user.get_full_name())

    @add_fixtures(['test_site'])
    def test_get_display_name_localsite_admin_private_other_site(self):
        """Testing Profile.get_display_name for a LocalSite admin viewing a
        member of another LocalSite with a private profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        viewer = User.objects.get(username='grumpy')
        site = LocalSite.objects.create(name='site-3')
        site.users.add(viewer)
        site.admins.add(viewer)

        self.assertEqual(profile.get_display_name(viewer), user.username)

    def test_get_display_name_self_private(self):
        """Testing Profile.get_display_name for a user viewing themselves with
        a private profile
        """
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True

        self.assertEqual(profile.get_display_name(user), user.get_full_name())



"""Extension hooks for augmenting e-mail messages."""

from __future__ import unicode_literals

import warnings
from collections import defaultdict
from inspect import getargspec

from reviewboard.deprecation import RemovedInReviewBoard40Warning
from reviewboard.reviews.signals import (review_request_published,
                                         review_published, reply_published,
                                         review_request_closed)


# A mapping of signals to EmailHooks.
_hooks = defaultdict(set)

# A mapping of signals to newly added arguments.
#
# See _call_hook_compat.
_hook_compat = defaultdict(list)
_hook_compat.update({
    review_published: ['to_owner_only'],
})


def register_email_hook(signal, handler):
    """Register an e-mail hook.

    Args:
        signal (django.dispatch.Signal):
            The signal that will trigger the e-mail to be sent. This is one of
            :py:data:`~reviewboard.reviews.signals.review_request_published`,
            :py:data:`~reviewboard.reviews.signals.review_request_closed`,
            :py:data:`~reviewboard.reviews.signals.review_published`, or
            :py:data:`~reviewboard.reviews.signals.reply_published`.

        handler (reviewboard.extensions.hooks.EmailHook):
            The ``EmailHook`` that will be triggered when an e-mail of the
            chosen type is about to be sent.
    """
    assert signal in (review_request_published, review_request_closed,
                      review_published, reply_published), (
        'Invalid signal %r' % signal)

    _hooks[signal].add(handler)


def unregister_email_hook(signal, handler):
    """Unregister an e-mail hook.

    Args:
        signal (django.dispatch.Signal):
            The signal that will trigger the e-mail to be sent. This is one of
            :py:data:`~reviewboard.reviews.signals.review_request_published`,
            :py:data:`~reviewboard.reviews.signals.review_request_closed`,
            :py:data:`~reviewboard.reviews.signals.review_published`, or
            :py:data:`~reviewboard.reviews.signals.reply_published`.

        handler (reviewboard.extensions.hooks.EmailHook):
            The ``EmailHook`` that will be triggered when an e-mail of the
            chosen type is about to be sent.
    """
    assert signal in (review_request_published, review_request_closed,
                      review_published, reply_published), (
        'Invalid signal %r' % signal)

    _hooks[signal].discard(handler)


def filter_email_recipients_from_hooks(to_field, cc_field, signal, **kwargs):
    """Filter the e-mail recipients through configured e-mail hooks.

    Args:
        to_field (set):
            The original To field of the e-mail, as a set of
            :py:class:`Users <django.contrib.auth.models.User>` and
            :py:class:`Groups <reviewboard.reviews.models.Group>`.

        cc_field (set):
            The original CC field of the e-mail, as a set of
            :py:class:`Users <django.contrib.auth.models.User>` and
            :py:class:`Groups <reviewboard.reviews.models.Group>`.

        signal (django.dispatch.Signal):
            The signal that triggered the e-mail.

        **kwargs (dict):
            Extra keyword arguments to pass to the e-mail hook.

    Returns:
        tuple:
        A 2-tuple of the To field and the CC field, as sets of
        :py:class:`Users <django.contrib.auth.models.User>` and
        :py:class:`Groups <reviewboard.reviews.models.Group>`.
    """
    optional_args = _hook_compat.get(signal)

    if signal in _hooks:
        for hook in _hooks[signal]:
            to_field = _call_hook_compat(
                hook,
                hook.get_to_field,
                optional_args=optional_args,
                value=to_field,
                **kwargs
            )
            cc_field = _call_hook_compat(
                hook,
                hook.get_cc_field,
                optional_args=optional_args,
                value=cc_field,
                **kwargs
            )

    return to_field, cc_field


def _call_hook_compat(hook, method, optional_args, value, **kwargs):
    """Call a hook in a backwards-compatible manner.

    This method allows newer versions of Review Board to add new arguments to
    EMail hooks without breaking backwards compatability. Instead, we inspect
    the method that would be called and if it does not accept any of the
    arguments listed in ``optional_args``, they will not be passed to the
    function and a warning will be emitted.

    Args:
        method (callable):
            The method on the
            :py:class:`reviewboard.extensions.hooks.EmailHook` class.

        optional_args (list of str):
            The names of the arguments that the hook may not have.

        value (object):
            The value to pass to the hook.

        **kwargs (dict):
            The keyword arguments that will be passed to ``method``.

    Returns:
        object:
        The return value of the passed method.
    """
    argspec = getargspec(method)
    removed = False

    if argspec.keywords is None:
        for arg in optional_args or []:
            if arg not in argspec.args:
                removed = True
                del kwargs[arg]

    if removed:
        warnings.warn(
            '%s.%s should accept **kwargs. Some arguments may not be passed '
            'to the hook.'
            % (hook.__name__, method.__name__),
            RemovedInReviewBoard40Warning
        )

    return method(value, **kwargs)



"""Forwarding imports and legacy utilities for authentication backends.

This module provides legacy functionality for registering authentication
backends, along with forwarding imports for:

.. autosummary::
   :nosignatures:

   ~reviewboard.accounts.backends.ad.ActiveDirectoryBackend
   ~reviewboard.accounts.backends.base.BaseAuthBackend
   ~reviewboard.accounts.backends.http_digest.HTTPDigestBackend
   ~reviewboard.accounts.backends.ldap.LDAPBackend
   ~reviewboard.accounts.backends.nis.NISBackend
   ~reviewboard.accounts.backends.registry.AuthBackendRegistry
   ~reviewboard.accounts.backends.registry.auth_backends
   ~reviewboard.accounts.backends.registry.get_enabled_auth_backends
   ~reviewboard.accounts.backends.standard.StandardAuthBackend
   ~reviewboard.accounts.backends.x509.X509Backend

Version Changed:
    3.0:
    The contents of this module were split into sub-modules.
"""

from __future__ import unicode_literals

from warnings import warn

from reviewboard.accounts.backends.ad import ActiveDirectoryBackend
from reviewboard.accounts.backends.base import BaseAuthBackend
from reviewboard.accounts.backends.http_digest import HTTPDigestBackend
from reviewboard.accounts.backends.ldap import LDAPBackend
from reviewboard.accounts.backends.nis import NISBackend
from reviewboard.accounts.backends.registry import (AuthBackendRegistry,
                                                    auth_backends,
                                                    get_enabled_auth_backends)
from reviewboard.accounts.backends.standard import StandardAuthBackend
from reviewboard.accounts.backends.x509 import X509Backend
from reviewboard.deprecation import RemovedInReviewBoard40Warning


# Legacy references.
AuthBackend = BaseAuthBackend
INVALID_USERNAME_CHAR_REGEX = AuthBackend.INVALID_USERNAME_CHAR_REGEX


def get_registered_auth_backends():
    """Yield all registered Review Board authentication backends.

    This will return all backends provided both by Review Board and by
    third parties that have properly registered with the
    ``reviewboard.auth_backends`` entry point.

    Deprecated:
        3.0:
        Iterate over
        :py:data:`~reviewboard.accounts.backends.registry.auth_backends`
        instead.

    Yields:
        type:
        The :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
        subclasses.
    """
    warn('reviewboard.accounts.backends.get_registered_auth_backends() is '
         'deprecated. Iterate over '
         'reviewboard.accounts.backends.auth_backends instead.',
         RemovedInReviewBoard40Warning)

    for backend in auth_backends:
        yield backend


def get_registered_auth_backend(backend_id):
    """Return the authentication backend with the specified ID.

    Deprecated:
        3.0:
        Use :py:meth:`auth_backends.get_auth_backend()
        <reviewboard.accounts.backends.registry.AuthBackendRegistry.
        get_auth_backend>` instead.

    Args:
        backend_id (unicode):
            The ID of the backend to retrieve.

    Returns:
        reviewboard.accounts.backends.base.BaseAuthBackend:
        The authentication backend, or ``None`` if it could not be found.
    """
    warn('reviewboard.accounts.backends.get_registered_auth_backend() is '
         'deprecated. Use '
         'reviewboard.accounts.backends.auth_backends.register() instead.',
         RemovedInReviewBoard40Warning)

    return auth_backends.get('backend_id', backend_id)


def register_auth_backend(backend_cls):
    """Register an authentication backend.

    This backend will appear in the list of available backends.
    The backend class must have a backend_id attribute set, and can only
    be registered once.

    Deprecated:
        3.0:
        Use :py:meth:`auth_backends.register()
        <reviewboard.accounts.backends.registry.AuthBackendRegistry.register>`
        instead.

    Args:
        backend_cls (type):
            The subclass of
            :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
            to register.

    Raises:
        KeyError:
            A backend already exists with this ID.
    """
    warn('reviewboard.accounts.backends.register_auth_backend() is '
         'deprecated. Use '
         'reviewboard.accounts.backends.auth_backends.register() instead.',
         RemovedInReviewBoard40Warning)

    auth_backends.register(backend_cls)


def unregister_auth_backend(backend_cls):
    """Unregister a previously registered authentication backend.

    Deprecated:
        3.0:
        Use :py:meth:`auth_backends.unregister()
        <reviewboard.accounts.backends.registry.AuthBackendRegistry.
        unregister>` instead.

    Args:
        backend_cls (type):
            The subclass of
            :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
            to unregister.
    """
    warn('reviewboard.accounts.backends.unregister_auth_backend() is '
         'deprecated. Use '
         'reviewboard.accounts.backends.auth_backends.unregister() instead.',
         RemovedInReviewBoard40Warning)

    auth_backends.unregister(backend_cls)


__all__ = (
    'ActiveDirectoryBackend',
    'AuthBackend',
    'AuthBackendRegistry',
    'HTTPDigestBackend',
    'INVALID_USERNAME_CHAR_REGEX',
    'LDAPBackend',
    'NISBackend',
    'StandardAuthBackend',
    'X509Backend',
    'auth_backends',
    'get_enabled_auth_backends',
    'get_registered_auth_backend',
    'get_registered_auth_backends',
    'register_auth_backend',
    'unregister_auth_backend',
)

__autodoc_excludes__ = (
    'ActiveDirectoryBackend',
    'AuthBackendRegistry',
    'HTTPDigestBackend',
    'LDAPBackend',
    'NISBackend',
    'StandardAuthBackend',
    'X509Backend',
    'auth_backends',
    'get_enabled_auth_backends',
)



"""Unit tests for reviewboard.accounts.mixins.CheckLoginRequiredViewMixin."""

from __future__ import unicode_literals

from django.contrib.auth.models import AnonymousUser, User
from django.http import HttpResponse, HttpResponseRedirect
from django.test.client import RequestFactory
from django.views.generic.base import View
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.accounts.mixins import CheckLoginRequiredViewMixin
from reviewboard.testing import TestCase


class CheckLoginRequiredViewMixinTests(TestCase):
    """Unit tests for reviewboard.accounts.mixins.CheckLoginRequiredViewMixin.
    """

    def test_dispatch_authenticated_user(self):
        """Testing CheckLoginRequiredViewMixin.dispatch with authenticated user
        """
        class MyView(CheckLoginRequiredViewMixin, View):
            def get(view, *args, **kwargs):
                self.assertTrue(view.request.user.is_authenticated())

                return HttpResponse('success')

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='doc',
                                                email='doc@example.com')

        view = MyView.as_view()
        response = view(request)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'success')

    def test_dispatch_anonymous_user_and_login_not_required(self):
        """Testing CheckLoginRequiredViewMixin.dispatch with anonymous user
        and login not required
        """
        class MyView(CheckLoginRequiredViewMixin, View):
            def get(view, *args, **kwargs):
                self.assertTrue(view.request.user.is_anonymous())

                return HttpResponse('success')

        self.siteconfig = SiteConfiguration.objects.get_current()
        self.siteconfig.set('auth_require_sitewide_login', False)
        self.siteconfig.save()

        request = RequestFactory().request()
        request.user = AnonymousUser()

        view = MyView.as_view()
        response = view(request)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'success')

    def test_dispatch_anonymous_user_and_login_required(self):
        """Testing CheckLoginRequiredViewMixin.dispatch with anonymous user
        and login required
        """
        class MyView(CheckLoginRequiredViewMixin, View):
            def get(view, *args, **kwargs):
                self.assertTrue(view.request.user.is_anonymous())

                return HttpResponse('success')

        self.siteconfig = SiteConfiguration.objects.get_current()
        self.siteconfig.set('auth_require_sitewide_login', True)
        self.siteconfig.save()

        request = RequestFactory().request()
        request.user = AnonymousUser()

        view = MyView.as_view()
        response = view(request)

        self.assertIsInstance(response, HttpResponseRedirect)



"""Tests for Review Board features."""

from __future__ import unicode_literals

from django.contrib.auth.models import AnonymousUser, User
from django.test.client import RequestFactory
from djblets.features import Feature, get_features_registry
from djblets.testing.decorators import add_fixtures

from reviewboard.features.checkers import RBFeatureChecker
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class DummyFeature(Feature):
    """A dummy feature for testing."""

    feature_id = 'dummy.feature'
    name = 'Dummy feature'
    summary = 'A dummy feature for testing.'


class RBFeatureCheckerTests(TestCase):
    """Tests for the RBFeatureChecker."""

    fixtures = ['test_site']

    FEATURE_ENABLED_SETTINGS = {
        DummyFeature.feature_id: True,
    }

    FEATURE_DISABLED_SETTINGS = {
        DummyFeature.feature_id: False,
    }

    @classmethod
    def setUpClass(cls):
        super(RBFeatureCheckerTests, cls).setUpClass()

        cls.request_factory = RequestFactory()

    def setUp(self):
        super(RBFeatureCheckerTests, self).setUp()
        self.feature = DummyFeature()
        self.local_site = LocalSite.objects.get(name='local-site-1')

    def tearDown(self):
        super(RBFeatureCheckerTests, self).tearDown()
        registry = get_features_registry()
        registry.unregister(self.feature)

    def test_local_site_feature_enabled(self):
        """Testing RBFeatureChecker.is_feature_enabled for a feature enabled on
        a local site
        """
        self.local_site.extra_data = {
            RBFeatureChecker.EXTRA_DATA_KEY: self.FEATURE_ENABLED_SETTINGS,
        }

        with self.settings(ENABLED_FEATURES=self.FEATURE_DISABLED_SETTINGS):
            self.assertTrue(RBFeatureChecker().is_feature_enabled(
                DummyFeature.feature_id, local_site=self.local_site))

    def test_local_site_feature_disabled(self):
        """Testing RBFeatureChecker.is_feature_enabled for a feature disabled
        on a local site
        """
        self.local_site.extra_data = {
            RBFeatureChecker.EXTRA_DATA_KEY: self.FEATURE_DISABLED_SETTINGS,
        }

        with self.settings(ENABLED_FEATURES=self.FEATURE_ENABLED_SETTINGS):
            self.assertTrue(RBFeatureChecker().is_feature_enabled(
                DummyFeature.feature_id, local_site=self.local_site))

    def test_local_site_feature_fallback_enabled(self):
        """Testing RBFeatureChecker.is_feature_enabled for an unconfigured
        feature on a LocalSite that is enabled globally
        """
        self.local_site.extra_data = {
            RBFeatureChecker.EXTRA_DATA_KEY: {},
        }

        with self.settings(ENABLED_FEATURES=self.FEATURE_ENABLED_SETTINGS):
            self.assertTrue(RBFeatureChecker().is_feature_enabled(
                DummyFeature.feature_id, local_site=self.local_site))

    def test_local_site_feature_fallback_disabled(self):
        """Testing RBFeatureChecker.is_feature_enabled for an unconfigured
        feature on a LocalSite that is disabled globally
        """
        self.local_site.extra_data = {
            RBFeatureChecker.EXTRA_DATA_KEY: {},
        }

        with self.settings(ENABLED_FEATURES=self.FEATURE_DISABLED_SETTINGS):
            self.assertFalse(RBFeatureChecker().is_feature_enabled(
                DummyFeature.feature_id, local_site=self.local_site))

    @add_fixtures(['test_users'])
    def test_local_site_feature_enabled_on_global(self):
        """Testing RBFeatureChecker.is_feature_enabled for a feature enabled on
        a local site while not on a site
        """
        self.local_site.extra_data = {
            RBFeatureChecker.EXTRA_DATA_KEY: self.FEATURE_ENABLED_SETTINGS,
        }
        self.local_site.save()

        request = self.request_factory.get('/')
        request.user = User.objects.get(username='doc')
        request.local_site = None

        with self.settings(ENABLED_FEATURES=self.FEATURE_DISABLED_SETTINGS):
            self.assertTrue(RBFeatureChecker().is_feature_enabled(
                DummyFeature.feature_id,
                request=request))

    @add_fixtures(['test_users'])
    def test_local_site_feature_disabled_on_global(self):
        """Testing RBFeatureChecker.is_feature_enabled for a feature disabled
        on a local site while not on a site
        """
        self.local_site.extra_data = {
            RBFeatureChecker.EXTRA_DATA_KEY: self.FEATURE_DISABLED_SETTINGS,
        }
        self.local_site.save()

        request = self.request_factory.get('/')
        request.user = User.objects.get(username='doc')
        request.local_site = None

        with self.settings(ENABLED_FEATURES=self.FEATURE_ENABLED_SETTINGS):
            self.assertTrue(RBFeatureChecker().is_feature_enabled(
                DummyFeature.feature_id,
                request=request))

    @add_fixtures(['test_users'])
    def test_cache_localsite_queries(self):
        """Testing RBFeatureChecker.is_feature_enabled caches LocalSite
        membership to reduce query count
        """
        self.local_site.extra_data = {
            RBFeatureChecker.EXTRA_DATA_KEY: self.FEATURE_ENABLED_SETTINGS,
        }
        self.local_site.save()

        request = self.request_factory.get('/')
        request.user = User.objects.get(username='doc')

        with self.assertNumQueries(1):
            for _ in range(3):
                self.assertTrue(RBFeatureChecker().is_feature_enabled(
                    DummyFeature.feature_id,
                    request=request))

    def test_no_queries_anonymous(self):
        """Testing RBFeatureChecker.is_feature_enabled does not query when the
        user is anonymous
        """
        self.local_site.extra_data = {
            RBFeatureChecker.EXTRA_DATA_KEY: self.FEATURE_ENABLED_SETTINGS,
        }
        self.local_site.save()

        request = self.request_factory.get('/')
        request.user = AnonymousUser()

        with self.assertNumQueries(0):
            for _ in range(3):
                self.assertFalse(RBFeatureChecker().is_feature_enabled(
                    DummyFeature.feature_id,
                    request=request))



"""Integrations support for Review Board.

This module provides the Review Board functionality needed to create
integrations for third-party services. It builds upon Djblets's integrations
foundation, offering some additional utilities for more easily creating
manageable integrations.

This module provides imports for:

.. autosummary::
   :nosignatures:

   ~reviewboard.integrations.base.get_integration_manager
   ~reviewboard.integrations.base.Integration
"""

from __future__ import unicode_literals

from reviewboard.integrations.base import get_integration_manager, Integration


__all__ = [
    'get_integration_manager',
    'Integration',
]



"""Review Board's extension resource."""

from djblets.extensions.resources import (
    ExtensionResource as DjbletsExtensionResource)
from djblets.util.decorators import augment_method_from

from reviewboard.extensions.base import get_extension_manager
from reviewboard.webapi.base import RBResourceMixin
from reviewboard.webapi.decorators import (webapi_check_local_site,
                                           webapi_check_login_required)


class ExtensionResource(RBResourceMixin, DjbletsExtensionResource):
    """Review Board's extension resource.

    This resource special-cases the one in Djblets to provide API token and
    OAuth token access.
    """

    @webapi_check_login_required
    @webapi_check_local_site
    @augment_method_from(DjbletsExtensionResource)
    def get(self, *args, **kwargs):
        pass

    @webapi_check_login_required
    @webapi_check_local_site
    @augment_method_from(DjbletsExtensionResource)
    def get_list(self, *args, **kwargs):
        pass

    @webapi_check_local_site
    @augment_method_from(DjbletsExtensionResource)
    def update(self, *args, **kwargs):
        pass

    def has_modify_permissions(self, request, extension, *args, **kwargs):
        """Return whether the user has access to modify this extension.

        Args:
            request (django.http.HttpRequest):
                The request.

            extension (reviewboard.extensions.base.Extension):
                The extension being modified.

            *args (tuple, unused):
                Additional positional arguments.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            bool:
            True, always. Individual permissions checks will be done via the
            :py:func:`djblets.webapi.decorators.webapi_permission_required`
            decorator.
        """
        return True



extension_resource = ExtensionResource(get_extension_manager())



"""Unit tests for reviewboard.reviews.views.PreviewReviewEmailView."""

from __future__ import unicode_literals

from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase


class PreviewReviewEmailViewTests(TestCase):
    """Unit tests for reviewboard.reviews.views.PreviewReviewEmailView."""

    fixtures = ['test_users', 'test_scmtools']

    def test_access_with_debug(self):
        """Testing PreviewReviewEmailView access with DEBUG=True"""
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request, publish=True)

        with self.settings(DEBUG=True):
            response = self.client.get(
                local_site_reverse(
                    'preview-review-email',
                    kwargs={
                        'review_request_id': review_request.pk,
                        'review_id': review.pk,
                        'message_format': 'text',
                    }))

        self.assertEqual(response.status_code, 200)

    def test_access_without_debug(self):
        """Testing PreviewReviewEmailView access with DEBUG=False"""
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request, publish=True)

        with self.settings(DEBUG=False):
            response = self.client.get(
                local_site_reverse(
                    'preview-review-email',
                    kwargs={
                        'review_request_id': review_request.pk,
                        'review_id': review.pk,
                        'message_format': 'text',
                    }))

        self.assertEqual(response.status_code, 404)

    def test_reply_access_with_debug(self):
        """Testing PreviewReviewEmailView with reply access and DEBUG=True"""
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request, publish=True)
        reply = self.create_reply(review, publish=True)

        with self.settings(DEBUG=True):
            response = self.client.get(
                local_site_reverse(
                    'preview-review-reply-email',
                    kwargs={
                        'review_request_id': review_request.pk,
                        'review_id': review.pk,
                        'reply_id': reply.pk,
                        'message_format': 'text',
                    }))

        self.assertEqual(response.status_code, 200)

    def test_reply_access_without_debug(self):
        """Testing PreviewReviewEmailView with reply access and DEBUG=False"""
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request, publish=True)
        reply = self.create_reply(review, publish=True)

        with self.settings(DEBUG=False):
            response = self.client.get(
                local_site_reverse(
                    'preview-review-reply-email',
                    kwargs={
                        'review_request_id': review_request.pk,
                        'review_id': review.pk,
                        'reply_id': reply.pk,
                        'message_format': 'text',
                    }))

        self.assertEqual(response.status_code, 404)



from __future__ import unicode_literals

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('WebAPIToken', 'policy', initial=None, null=True),
    ChangeField('WebAPIToken', 'extra_data', initial=None, null=True),
]
